import asyncio
from unittest import TestCase
from unittest.mock import Mock

import time
from pycoin.message.InvItem import ITEM_TYPE_TX

from spruned.daemon.bitcoin_p2p.p2p_connection import P2PConnection
from test.utils import async_coro


class TestP2PConnection(TestCase):
    def setUp(self):
        self.loopmock = Mock()
        self.online_checker = Mock()
        self.delayer = Mock()
        self._peer_name = ('127.0.0.1', 8333, 'peer_name')
        self.peer_factory = Mock()
        self.peer_factory()._reader = Mock(_transport=Mock())
        self.peer_factory()._reader._transport.get_extra_info.return_value = self._peer_name
        self.connector = Mock()
        self.hostname = 'hostname'
        self.port = 8333
        self.sut = P2PConnection(
            self.hostname, self.port, loop=self.loopmock,
            is_online_checker=self.online_checker, delayer=self.delayer,
            call_timeout=3, connector=self.connector, peer=self.peer_factory
        )
        self.loop = asyncio.get_event_loop()

    def test_connect_ok(self):
        callback = Mock()
        callback_blocks = Mock()
        self.sut.add_on_blocks_callback(callback_blocks)
        self.sut.add_on_connect_callback(callback)
        reader, writer = Mock(), Mock()
        self.peer_factory().perform_handshake.return_value = async_coro('the peer version answer')
        self.connector.return_value = async_coro((reader, writer))
        self.peer_factory().next_message.return_value = async_coro(None)
        self.loop.run_until_complete(self.sut.connect())
        self.assertEqual(self.sut._version, 'the peer version answer')
        self.assertEqual(len(self.loopmock.method_calls), 1)
        Mock.assert_not_called(callback_blocks)
        self.assertTrue(self.sut.connected)

    def test_connect_exception(self):
        self.connector.side_effect = ValueError
        self.loop.run_until_complete(self.sut.connect())
        self.assertIsNone(self.sut._version)

    def test_disconnect(self):
        self.loop.run_until_complete(self.sut.disconnect())
        peer = Mock()
        self.sut.peer = peer
        self.sut.peer.close.return_value = True
        self.loop.run_until_complete(self.sut.disconnect())
        Mock.assert_called_once_with(peer.close)
        self.assertIsNone(self.sut.peer)

    def test_on_events(self):
        called = False

        async def txs_callback(*a, **kw):
            nonlocal called
            called = True

        async def wait():
            nonlocal called
            now = int(time.time())
            while not called:
                await asyncio.sleep(2)
                if int(time.time()) - now > 5:
                    raise AssertionError('Timeout!')

        async def start():
            self.sut._on_inv('inv', 'tx', {'items': [item]})

        item = Mock(item_type=ITEM_TYPE_TX)
        self.sut.loop = self.loop
        self.sut.add_on_transaction_callback(txs_callback)
        self.loop.run_until_complete(asyncio.gather(
            wait(), start()
        ))
        self.sut.loop = self.loopmock
        self.assertTrue(called)

    def test_on_ping(self):
        self.sut.peer = Mock()
        self.sut._on_ping('ping', 'ping', {'nonce': 'cafe'})
        Mock.assert_called_once_with(self.sut.peer.send_msg, 'pong', nonce='cafe')



import asyncio
import pickle
import time

from spruned.application.database import ldb_batch
from spruned.application.logging_factory import Logger
from spruned.application.tools import async_delayed_task
from spruned.repositories.blockchain_repository import TRANSACTION_PREFIX, BLOCK_PREFIX


class CacheAgent:
    def __init__(self, repository, limit, loop=asyncio.get_event_loop(), delayer=async_delayed_task):
        self.session = repository.ldb
        self.repository = repository
        self.repository.blockchain.set_cache(self)
        self.cache_name = b'cache_index'
        self.index = None
        self.limit = limit
        self._last_dump_size = None
        self.loop = loop
        self.lock = asyncio.Lock()
        self.delayer = delayer

    def init(self):
        self._load_index()

    def dump(self):
        self._save_index()

    def _deserialize_index(self, rawdata):
        index = {'keys': {}}
        data = pickle.loads(rawdata)
        s = 0
        for d in data:
            s += d[2]
            index['keys'][d[0]] = {
                'saved_at': d[1],
                'size': d[2],
                'key': d[0]
            }
        index['total'] = s
        self.index = index
        return index

    def _serialize_index(self):
        data = []
        sorted_data = sorted(self.index['keys'].values(), key=lambda x: x['saved_at'])
        for x in sorted_data:
            data.append([x['key'], x['saved_at'], x['size']])
        return pickle.dumps(data)

    @ldb_batch
    def _save_index(self):
        data = self._serialize_index()
        self.session.put(self.cache_name, data)
        Logger.cache.debug('Saved index')
        self._last_dump_size = self.index['total']

    def _load_index(self):
        index = self.session.get(self.cache_name)
        if not index:
            Logger.cache.warning('Cache not found. Ok if is the first time')
            return
        Logger.cache.debug('Loaded index')
        index and self._deserialize_index(index)
        self._last_dump_size = self.index and self.index['total']

    def track(self, key, size):
        if not self.index:
            self.index = {'keys': {}, 'total': 0}
        self.index['keys'][key] = {
            'saved_at': int(time.time()),
            'size': size,
            'key': key
        }
        self.index['total'] += size

    async def check(self):
        if self.index and self.index.get('keys') and not self._last_dump_size:
            Logger.cache.info('Pending data, dumping')
            self._save_index()
        if not self.index:
            Logger.cache.info('No prev index found, trying to load')
            self._load_index()
        if not self.index:
            self.index = {'keys': {}, 'total': 0}
            Logger.cache.info('No prev index found nor loaded')
            self._save_index()
            return
        else:
            self._purge_stales()
        if self.index['total'] > self.limit:
            Logger.cache.info('Purging cache, size: %s, limit: %s', self.index['total'], self.limit)
            blockfirst = {0: 2, 1: 1}
            index_sorted = sorted(
                self.index['keys'].values(), key=lambda x: ((blockfirst[x['key'][0]] ** 33) + x['saved_at'])
            )
            i = 0
            while self.index['total'] * 1.1 > self.limit:
                item = index_sorted[i]
                Logger.cache.debug('Deleting %s' % item)
                self.delete(item)
                i += 1
        else:
            Logger.cache.info('Cache is ok, size: %s, limit: %s', self.index['total'], self.limit)
        if self.index['total'] != self._last_dump_size:
            self._save_index()

    def delete(self, item):
        if item['key'][0] == int.from_bytes(BLOCK_PREFIX, 'little'):
            Logger.leveldb.debug('Deleting block %s', item)
            self.repository.blockchain.remove_block(item['key'][2:])
        else:
            raise ValueError('Problem: %s' % item)
        self.index['total'] -= self.index['keys'].pop(item['key'])['size']

    async def lurk(self):
        try:
            await self.lock.acquire()
            await self.check()
        finally:
            self.lock.release()
            self.loop.create_task(self.delayer(self.lurk(), 600))

    def get_index(self):
        if not self.index:
            self._load_index()
        return self.index

    def _purge_stales(self):
        stales = []
        for key in self.index['keys']:
            if not self.session.get(self.repository.blockchain.storage_name + b'.' + key):
                stales.append(key)
        for stale in stales:
            self.index['total'] -= self.index['keys'].pop(stale)['size']
        Logger.cache.debug('Stales purge done, removed %s items from index', len(stales))



import asyncio
import unittest
from unittest.mock import Mock
import binascii

import time

from spruned.application.tools import async_delayed_task
from spruned.daemon.electrod.electrod_connection import ElectrodConnection
from test.utils import async_coro, coro_call


class TestElectrodConnection(unittest.TestCase):

    def setUp(self):
        self.checker = Mock().checker
        self.checker.return_value = True
        self.client = Mock()
        self.client_factory = lambda: self.client
        self.serverinfo = Mock()
        self.electrod_loop = Mock()
        self.delayer = Mock()
        self.sut = ElectrodConnection(
            'hostname', 's', client=self.client_factory, expire_errors_after=3,
            is_online_checker=self.checker, serverinfo=self.serverinfo, loop=self.electrod_loop,
            delayer=self.delayer
        )
        self.assertEqual(self.sut.start_score, self.sut._start_score)
        self.loop = asyncio.get_event_loop()

    def tearDown(self):
        self.client.reset_mock()

    def test_errors_score(self):
        self.loop.run_until_complete(
            asyncio.gather(
                self.sut.on_error('error1'),
                self.sut.on_error('error1'),
                self.sut.on_error('error1')
            )
        )
        self.assertEqual(self.sut.score, 7)
        time.sleep(3.01)
        self.assertEqual(self.sut.score, 10)

    def test_connect_success(self):
        self.serverinfo.return_value = 'server info'
        self.client.connect.return_value = async_coro(None)
        self.client.server_version = 'ElectrumX 1.2'
        callback = Mock()
        callback.return_value = async_coro(True)
        self.sut.add_on_connect_callback(callback)

        self.loop.run_until_complete(self.sut.connect())
        self.assertTrue(self.sut.version, 'ElectrumX 1.2')
        Mock.assert_called_once_with(
            self.client.connect,
            'server info',
            disconnect_callback=self.sut.on_connectrum_disconnect,
            disable_cert_verify=True,
            use_tor=self.sut.use_tor
        )
        Mock.assert_called_once_with(
            self.serverinfo,
            self.sut.nickname, hostname='hostname', ports='s', version='1.2'
        )
        self.assertEqual(8, len(binascii.unhexlify(self.sut.nickname)))
        self.assertEqual(callback.call_count, 1)

    def test_disconnection(self):
        self.client.protocol = True
        self.client.close.side_effect = ConnectionError
        self.loop.run_until_complete(self.sut.disconnect())
        self.assertFalse(self.sut.connected)
        Mock.assert_called_once_with(self.client.close)

    def test_connect_failure(self):
        called = False

        async def connfalse(peer, error_type=None):
            nonlocal called
            called = True
            self.assertEqual(error_type, 'connect')
            self.checker.return_value = False

        self.client.connect.side_effect = ConnectionError
        self.assertTrue(self.sut.is_online())
        self.sut.add_on_error_callback(connfalse)
        self.sut.loop = self.loop
        self.loop.run_until_complete(self.sut.connect())
        self.assertIsNone(self.sut._version)
        self.assertTrue(called)
        self.assertFalse(self.sut.is_online())

    def test_connection_timeout(self):
        self.delayer.return_value = 'delayed'
        self.client.connect.return_value = asyncio.sleep(2)
        self.sut._timeout = 1
        self.loop.run_until_complete(self.sut.connect())
        self.assertIsNone(self.sut._version)
        self.assertEqual(1, len(self.sut.errors))
        self.assertEqual(9, self.sut.score)

    def test_rpc_call_success(self):
        self.client.RPC.return_value = async_coro(True)
        res = self.loop.run_until_complete(self.sut.rpc_call('method', ('cafe', 'babe')))
        Mock.assert_called_once_with(
            self.client.RPC,
            'method', 'cafe', 'babe'
        )
        self.assertEqual(res, True)

    def test_rpc_call_error(self):
        self.delayer.return_value = 'delayed'
        self.client.RPC.return_value = ConnectionError
        res = self.loop.run_until_complete(self.sut.rpc_call('method', ('cafe', 'babe')))
        self.assertEqual(res, None)
        Mock.assert_called_once_with(self.electrod_loop.create_task, 'delayed')
        Mock.assert_called_once_with(self.delayer, coro_call('on_error'))

    def test_ping_success(self):
        self.client.RPC.return_value = asyncio.gather(asyncio.sleep(1.1), async_coro('ElectrumX 1.2'))
        res = self.loop.run_until_complete(self.sut.ping())
        self.assertEqual(1.1 < res < 1.15, True)

    def test_ping_timeout(self):
        self.client.RPC.return_value = asyncio.sleep(4)
        res = self.loop.run_until_complete(self.sut.ping(timeout=1))
        self.assertIsNone(res)

    def test_subscribe_and_fail(self):
        queue = Mock()
        queue.get.side_effect = [async_coro([{'2nd': 'header'}]), ConnectionError]
        future = {'1st': 'header'}
        self.client.subscribe.return_value = async_coro(future), queue
        sub_called = False

        async def on_subscription(*a):
            nonlocal sub_called
            self.assertEqual(a[0]._last_header, {'1st': 'header'})
            sub_called = True

        self.sut.loop = self.loop
        self.sut.delayer = async_delayed_task
        self.loop.run_until_complete(self.sut.subscribe('channel', on_subscription, lambda *a: async_coro(True)))
        Mock.assert_not_called(self.electrod_loop.create_task)
        self.assertTrue(sub_called)
        self.sut._last_header = {'2nd': 'header'}

    def test_subscribe_error(self):
        self.sut.loop = self.electrod_loop
        self.delayer.return_value = 'delayer'
        self.client.subscribe.side_effect = ConnectionError
        self.loop.run_until_complete(self.sut.subscribe('cafe', 'babe', 'c0ca'))
        Mock.assert_called_with(self.electrod_loop.create_task, 'delayer')
        Mock.assert_called_with(self.delayer, coro_call('on_error'))

    def test_callbacks(self):
        hcb = Mock(return_value=async_coro(True))
        ccb = Mock(return_value=async_coro(True))
        dcb = Mock(return_value=async_coro(True))
        pcb = Mock(return_value=async_coro(True))
        ecb = Mock(return_value=async_coro(True))
        self.sut.add_on_header_callbacks(hcb)
        self.sut.add_on_error_callback(ecb)
        self.sut.add_on_connect_callback(ccb)
        self.sut.add_on_disconnect_callback(dcb)
        self.sut.add_on_peers_callback(pcb)
        self.loop.run_until_complete(
            asyncio.gather(
                self.sut.on_header('header'),
                self.sut.on_connect(),
                self.sut.on_error('error'),
                self.sut.on_peers()
            )
        )
        self.sut.on_connectrum_disconnect()
        for m in [hcb, ccb, dcb, pcb]:
            Mock.assert_called_with(m, self.sut)
        Mock.assert_called_with(ecb, self.sut, error_type='error')
        self.assertEqual(self.sut.last_header, 'header')



import asyncio
import binascii
from typing import Dict
from spruned.application.context import ctx
from spruned.application.exceptions import InvalidPOWException
from spruned.application.logging_factory import Logger
from spruned.daemon import exceptions
from spruned.application.tools import blockheader_to_blockhash, deserialize_header, serialize_header
from spruned.daemon.electrod.electrod_connection import ElectrodConnectionPool, ElectrodConnection


class ElectrodInterface:
    def __init__(self, connectionpool: ElectrodConnectionPool, loop=asyncio.get_event_loop()):
        self._network = ctx.get_network()
        self.pool = connectionpool
        self._checkpoints = self._network['checkpoints']
        self.loop = loop

    @property
    def is_pool_online(self):  # pragma: no cover
        return self.pool.is_online()

    def _parse_header(self, electrum_header: Dict):
        header_hex = serialize_header(electrum_header)
        blockhash_from_header = blockheader_to_blockhash(header_hex)
        if electrum_header['block_height'] in self._checkpoints:
            if self._checkpoints[electrum_header['block_height']] != blockhash_from_header:
                raise exceptions.NetworkHeadersInconsistencyException

        header_data = deserialize_header(header_hex)
        return {
            'block_hash': blockhash_from_header,
            'block_height': electrum_header['block_height'],
            'header_bytes': binascii.unhexlify(header_hex),
            'prev_block_hash': header_data['prev_block_hash'],
            'timestamp': header_data['timestamp']
        }

    def add_header_subscribe_callback(self, callback):
        async def parse_and_go(peer, res):
            try:
                header = self._parse_header(res)
                return await callback(peer, header)
            except InvalidPOWException:
                Logger.electrum.error('Wrong POW for header %s from peer %s. Banning', res, peer)
                self.loop.create_task(peer.disconnect())
        self.pool.add_header_observer(parse_and_go)

    def add_on_connected_callback(self, callback):
        self.pool.add_on_connected_observer(callback)

    async def get_header(self, height: int, fail_silent_out_of_range=False, get_peer=False):
        try:
            response = await self.pool.call(
                'blockchain.block.get_header', height, get_peer=True
            )
            peer, header = response
        except exceptions.ElectrodMissingResponseException:
            if fail_silent_out_of_range:
                return
            raise
        try:
            parsed_header = self._parse_header(header)
        except (exceptions.NetworkHeadersInconsistencyException, InvalidPOWException):
            Logger.electrum.error('Wrong POW for header %s from peer %s. Banning', header, peer)
            self.loop.create_task(peer.disconnect())
            return

        if not get_peer:
            return parsed_header
        return peer, parsed_header

    async def handle_peer_error(self, peer):
        await self.pool.on_peer_error(peer)

    async def getrawtransaction(self, txid: str, verbose=False):
        if txid == self._network['tx0']:
            raise exceptions.GenesisTransactionRequestedException
        return await self.pool.call('blockchain.transaction.get', txid)

    async def listunspents_by_address(self, address: str):
        return await self.pool.call('blockchain.address.listunspent', address)

    async def listunspents_by_scripthash(self, scripthash: str, get_peer=False, fail_silent=False):
        return await self.pool.call(
            'blockchain.scripthash.listunspent', scripthash, get_peer=get_peer, fail_silent=fail_silent
        )

    async def getaddresshistory(self, scripthash: str):
        return await self.pool.call('blockchain.address.get_history', scripthash)

    async def get_chunk(self, chunks_index: int, get_peer=False):
        print('get chunk: chunks index %s ' % chunks_index)
        return await self.pool.call('blockchain.block.get_chunk', chunks_index, get_peer=get_peer)

    async def get_merkleproof(self, txid: str, block_height: int):
        return await self.pool.call('blockchain.transaction.get_merkle', txid, block_height)

    async def get_headers_in_range_from_chunks(self, starts_from: int, ends_to: int, get_peer=False):
        futures = []
        for chunk_index in range(starts_from, ends_to):
            futures.append(self.get_headers_from_chunk(chunk_index, get_peer=get_peer))
        headers = []
        if not get_peer:
            for _headers in await asyncio.gather(*futures):
                _headers and headers.extend(_headers)
            return headers
        else:
            peer = None
            for response in await asyncio.gather(*futures):
                response and headers.extend(response[1])
                peer = response[0]
            return peer, headers

    async def get_headers_in_range(self, starts_from: int, ends_to: int):
        chunks_range = [x for x in range(starts_from, ends_to)]
        futures = []
        for i in chunks_range:
            futures.append(self.get_header(i))
        return await asyncio.gather(*futures)

    async def estimatefee(self, blocks: int):
        return await self.pool.call('blockchain.estimatefee', blocks)

    async def get_headers_from_chunk(self, chunk_index: int, get_peer=True):
        peer = None
        if get_peer:
            res = await self.get_chunk(chunk_index, get_peer=get_peer)
            peer, chunk = res if res else (None, None)
        else:
            chunk = await self.get_chunk(chunk_index, get_peer=get_peer)
        if not chunk:
            return
        hex_headers = [chunk[i:i + 160] for i in range(0, len(chunk), 160)]
        headers = []
        for i, header_hex in enumerate(hex_headers):
            header = deserialize_header(header_hex)
            header['block_height'] = int(chunk_index * 2016 + i)
            header['header_bytes'] = binascii.unhexlify(header_hex)
            header['block_hash'] = header.pop('hash')
            if header['block_height'] in self._checkpoints:
                if self._checkpoints[header['block_height']] != header['block_hash']:
                    await peer.disconnect()
                    raise exceptions.NetworkHeadersInconsistencyException(
                        'Checkpoint failure. Expected: %s, Failure: %s',
                        self._checkpoints[header['block_height']], header['block_hash']
                    )
            headers.append(header)
        return get_peer and (peer, headers) or headers

    async def start(self):
        self.loop.create_task(self.pool.connect())

    async def disconnect_from_peer(self, peer: ElectrodConnection):
        self.loop.create_task(peer.disconnect())

    async def sendrawtransaction(self, rawtx: str):
        return await self.pool.call('blockchain.transaction.broadcast', rawtx)



class SprunedException(Exception):
    pass


class ServiceException(SprunedException):
    pass


class HTTPClientException(SprunedException):
    pass


class SpentTxOutException(SprunedException):
    pass


class InvalidPOWException(SprunedException):
    pass


class SourcesDisagreementException(SprunedException):
    pass



from spruned.application.exceptions import SprunedException


class HeadersInconsistencyException(SprunedException):
    pass


class NoQuorumOnResponsesException(SprunedException):
    pass


class ConsistencyCheckRetryException(SprunedException):
    pass


class NoPeersException(SprunedException):
    pass


class NetworkHeadersInconsistencyException(SprunedException):
    pass


class NoHeadersException(SprunedException):
    pass


class ElectrodMissingResponseException(SprunedException):
    pass


class NoServersException(SprunedException):
    pass


class BlocksInconsistencyException(SprunedException):
    pass


class GenesisTransactionRequestedException(SprunedException):
    pass



import abc
from typing import List


class ConnectionAbstract(metaclass=abc.ABCMeta):  # pragma: no cover

    @property
    @abc.abstractmethod
    def start_score(self):
        pass

    @property
    @abc.abstractmethod
    def hostname(self):
        pass

    @property
    @abc.abstractmethod
    def score(self):
        pass

    @property
    @abc.abstractmethod
    def version(self):
        pass

    @property
    @abc.abstractmethod
    def connected(self):
        pass

    @property
    @abc.abstractmethod
    def last_header(self):
        pass

    @abc.abstractmethod
    def is_online(self) -> bool:
        pass

    @abc.abstractmethod
    def add_on_header_callbacks(self, callback):
        pass

    @abc.abstractmethod
    def add_on_connect_callback(self, callback):
        pass

    @abc.abstractmethod
    def add_on_disconnect_callback(self, callback):
        pass

    @abc.abstractmethod
    def add_on_peers_callback(self, callback):
        pass

    @abc.abstractmethod
    def connect(self):
        pass

    @abc.abstractmethod
    def ping(self, timeout=None):
        pass

    @property
    @abc.abstractmethod
    def errors(self):
        pass

    @abc.abstractmethod
    def disconnect(self):
        pass

    @abc.abstractmethod
    def add_error(self, *a):
        pass


class ConnectionPoolAbstract(metaclass=abc.ABCMeta):  # pragma: no cover
    @property
    @abc.abstractmethod
    def established_connections(self) -> List:
        pass

    @property
    @abc.abstractmethod
    def connections(self) -> List:
        pass

    @abc.abstractmethod
    def add_on_connected_observer(self, observer):
        pass

    @abc.abstractmethod
    def add_header_observer(self, observer):
        pass

    @abc.abstractmethod
    async def connect(self):
        pass

    @abc.abstractmethod
    def is_online(self) -> bool:
        pass



from spruned.application.context import ctx as _ctx, Context


def builder(ctx: Context):  # pragma: no cover
    from spruned.application.cache import CacheAgent
    from spruned.repositories.repository import Repository
    from spruned.daemon.tasks.blocks_reactor import BlocksReactor
    from spruned.daemon.tasks.headers_reactor import HeadersReactor
    from spruned.application import spruned_vo_service
    from spruned.application.jsonrpc_server import JSONRPCServer
    from spruned.daemon.electrod import build as electrod_builder
    from spruned.daemon.bitcoin_p2p import build as p2p_builder

    electrod_connectionpool, electrod_interface = electrod_builder(ctx.get_network())
    p2p_connectionpool, p2p_interface = p2p_builder(ctx.get_network())

    repository = Repository.instance()
    cache = CacheAgent(repository, int(ctx.cache_size))
    repository.set_cache(cache)

    service = spruned_vo_service.SprunedVOService(
        electrod_interface,
        p2p_interface,
        repository=repository,
        cache=cache
    )
    jsonrpc_server = JSONRPCServer(ctx.rpcbind, ctx.rpcport, ctx.rpcuser, ctx.rpcpassword)
    jsonrpc_server.set_vo_service(service)
    headers_reactor = HeadersReactor(repository.headers, electrod_interface)
    blocks_reactor = BlocksReactor(repository, p2p_interface, prune=int(ctx.keep_blocks))
    return jsonrpc_server, headers_reactor, blocks_reactor, repository, cache


jsonrpc_server, headers_reactor, blocks_reactor, repository, cache = builder(_ctx)




async def async_coro(response):
    return response


class MockValidator(object):

    def __init__(self, validator):
        # validator is a function that takes a single argument and returns a bool.
        self.validator = validator

    def __eq__(self, other):
        return bool(self.validator(other))


def coro_call(coro_name: str):
    def test_coro(coro):
        return coro.cr_code.co_name == coro_name

    return MockValidator(test_coro)


def in_range(n1: int, n2: int):
    def test(n):
        return n1 <= n <= n2
    return MockValidator(test)


def make_headers(s: int, e: int, prevblock_0: str= None, with_timestamp=True):
    import os
    import binascii
    _headers = []
    for i, x in enumerate(range(s, e)):
        block_hash = binascii.hexlify(os.urandom(32)).decode()
        _headers.append({
            'block_hash': block_hash,
            'prev_block_hash': x != 0 and (_headers[i-1]['block_hash'] if i > 0 else prevblock_0),
            'timestamp': 1,
            'header_bytes': os.urandom(80),
            'block_height': x
        }
        )
        if x == 0:
            _headers[0].pop('prev_block_hash')
        if not with_timestamp:
            _headers[i].pop('timestamp')
    return _headers



from argparse import Namespace
import threading
from pathlib import Path
from typing import Dict
from spruned.application import networks


class Context(dict):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.configfile = kw.get('configfile', 'spruned.conf')
        self.update(
            {
                'configfile': {},
                'args': {},
                'default': {
                    'daemonize': False,
                    'datadir': str(Path.home()) + '/.spruned',
                    'rpcbind': '127.0.0.1',
                    'rpcport': None,
                    'rpcuser': 'rpcuser',
                    'rpcpassword': 'rpcpassword',
                    'network': 'bitcoin.mainnet',
                    'debug': False,
                    'cache_size': 50,
                    'keep_blocks': 200
                }
            }
        )
        self.load_config()

    def load_config(self):
        values = {
            'i': ['cache_size', 'keep_blocks', 'rpcport'],
            'b': ['daemonize', 'debug']
        }
        import os
        filename = self.datadir + '/' + self.configfile
        if not os.path.exists(filename):
            return
        with open(filename, 'r') as f:
            lines = f.readlines()
        for i, line in enumerate(lines, 1):
            line = line.strip().replace(' ', '')
            if not line:
                continue
            k, v = line.split('=')
            if k not in self['default']:
                raise ValueError('Configuration file error: parameter not admitted: %s (%s:%s)' % (line, filename, i))
            if k in values['i']:
                self['configfile'][k] = int(v)
            elif k in values['b']:
                self['configfile'][k] = bool(v)
            else:
                self['configfile'][k] = v

    @property
    def datadir(self):
        if self._get_param('network') != 'bitcoin.mainnet':
            return self._get_param('datadir') + '/' + self._get_param('network')
        return self._get_param('datadir')

    @property
    def debug(self):
        return self._get_param('debug')

    @property
    def keep_blocks(self):
        return int(self._get_param('keep_blocks'))

    @property
    def network(self):
        return self._get_param('network')

    @property
    def rpcbind(self):
        return self._get_param('rpcbind')

    @property
    def rpcport(self):
        return self._get_param('rpcport') or self.get_network().get('rpc_port')

    @property
    def rpcuser(self):
        return self._get_param('rpcuser')

    @property
    def rpcpassword(self):
        return self._get_param('rpcpassword')

    @property
    def daemonize(self):
        return self._get_param('daemonize')

    @property
    def cache_size(self):
        return int(self._get_param('cache_size'))

    def load_args(self, args: Namespace):
        self['args'] = {
            'daemonize': args.daemonize,
            'datadir': args.datadir,
            'rpcbind': args.rpcbind,
            'rpcpassword': args.rpcpassword,
            'rpcport': args.rpcport,
            'rpcuser': args.rpcuser,
            'network': args.network,
            'debug': args.debug,
            'cache_size': int(args.cache_size),
            'keep_blocks': int(args.keep_blocks)
        }
        self.apply_context()

    def _get_param(self, key):
        return self['args'].get(key, None) or \
               self['configfile'].get(key, None) or \
               self['default'].get(key, None)

    def apply_context(self):
        pass

    def get_network(self) -> Dict:
        net, work = self._get_param('network').split('.')
        module = getattr(networks, net)
        return getattr(module, work)


_local = threading.local()
_local.ctx = ctx = Context()



import asyncio
import random
from unittest import TestCase
from unittest.mock import Mock, call
from spruned.application.jsonrpc_server import JSONRPCServer
from spruned.application.utils.jsonrpc_client import JSONClient
from test.utils import async_coro


class TestJSONRPCServerSendrawtransaction(TestCase):
    def setUp(self):
        bindport = random.randint(31337, 41337)
        self.sut = JSONRPCServer('127.0.0.1', bindport, 'testuser', 'testpassword')
        self.vo_service = Mock()
        self.sut.set_vo_service(self.vo_service)
        self.client = JSONClient(b'testuser', b'testpassword', '127.0.0.1', bindport)
        self.loop = asyncio.get_event_loop()

    def test_sendrawtransaction_success(self):
        self.vo_service.sendrawtransaction.side_effect = [
            async_coro('bedf6b43cf9a9278b8637d89f378a8f25b8f0e1be729325f7eea74a3096d9520')
            ]

        async def test():
            await self.sut.start()
            response = await self.client.call('sendrawtransaction', params=['cafebabe'])
            return response

        res = self.loop.run_until_complete(test())
        self.assertEqual(
            res,
            {
                'error': None, 'id': 1, 'jsonrpc': '2.0',
                'result': 'bedf6b43cf9a9278b8637d89f378a8f25b8f0e1be729325f7eea74a3096d9520'
             }
        )

    def test_getrawtransaction_error_decode(self):
        async def test():
            await self.sut.start()
            response = await self.client.call('sendrawtransaction', params=["nonhex"])
            return response

        res = self.loop.run_until_complete(test())
        self.assertEqual(
            res,
            {'error': {'code': -22, 'message': 'TX decode failed'},
             'id': 1,
             'jsonrpc': '2.0',
             'result': None}
        )
        Mock.assert_not_called(self.vo_service.getblock)


import asyncio
from unittest import TestCase
from unittest.mock import Mock, create_autospec, call

from spruned.daemon.bitcoin_p2p.p2p_interface import P2PInterface
from spruned.daemon.tasks.blocks_reactor import BlocksReactor
from spruned.repositories.repository import Repository
from test.utils import async_coro


class TestBlocksReactory(TestCase):
    def setUp(self):
        self.interface = create_autospec(P2PInterface)
        self.repo = create_autospec(Repository)
        self.loopmock = Mock()
        self.delayer = Mock()
        self.sut = BlocksReactor(self.repo, self.interface, self.loopmock, prune=5, delayed_task=self.delayer)
        self.loop = asyncio.get_event_loop()

    def test_check_blockchain_local_behind_remote(self):
        self.sut.set_last_processed_block({'block_hash': 'cafe', 'block_height': 9})
        self.repo.headers.get_best_header.return_value = {'block_hash': 'babe', 'block_height': 10}
        self.interface.get_blocks.return_value = async_coro({'babe': {'block_hash': 'babe', 'block_bytes': b'raw'}})
        self.repo.headers.get_headers_since_height.return_value = [{'block_hash': 'babe', 'block_height': 10}]
        self.repo.blockchain.get_block.return_value = None
        self.repo.blockchain.save_blocks.side_effect = lambda *x: x
        self.loop.run_until_complete(self.sut.check())
        Mock.assert_called_once_with(
            self.repo.blockchain.save_blocks, {'block_hash': 'babe', 'block_bytes': b'raw'}
        )
        self.assertEqual(self.sut._last_processed_block, {'block_hash': 'babe', 'block_height': 10})
        Mock.assert_called_once_with(self.repo.headers.get_best_header)
        Mock.assert_called_once_with(self.interface.get_blocks, 'babe')
        Mock.assert_called_once_with(self.repo.headers.get_headers_since_height, 9, limit=10)
        Mock.assert_called_once_with(self.repo.blockchain.get_block, 'babe', with_transactions=False)
        Mock.assert_called_once_with(self.repo.blockchain.save_blocks, {'block_hash': 'babe', 'block_bytes': b'raw'})

    def test_check_blockchain_local_behind_remote_but_block_already_stored(self):
        self.sut.set_last_processed_block({'block_hash': 'cafe', 'block_height': 9})
        self.repo.headers.get_best_header.return_value = {'block_hash': 'babe', 'block_height': 10}
        self.interface.get_blocks.return_value = async_coro({'babe': {'block_hash': 'babe', 'block_bytes': b'raw'}})
        self.repo.headers.get_headers_since_height.return_value = [{'block_hash': 'babe', 'block_height': 10}]
        self.repo.blockchain.get_block.return_value = {'block_hash': 'babe', 'block_bytes': b'raw'}
        self.repo.blockchain.save_blocks.side_effect = lambda *x: x
        self.loop.run_until_complete(self.sut.check())
        Mock.assert_called_once_with(self.repo.headers.get_best_header)
        Mock.assert_called_once_with(self.repo.headers.get_headers_since_height, 9, limit=10)
        Mock.assert_called_once_with(self.repo.blockchain.get_block, 'babe', with_transactions=False)
        Mock.assert_not_called(self.interface.get_blocks)
        Mock.assert_not_called(self.repo.blockchain.save_blocks)
        self.assertEqual(self.sut._last_processed_block, {'block_hash': 'babe', 'block_height': 10})

    def test_check_blockchain_local_behind_remote_error_saving_block(self):
        self.sut.set_last_processed_block({'block_hash': 'cafe', 'block_height': 9})
        self.repo.headers.get_best_header.return_value = {'block_hash': 'babe', 'block_height': 10}
        self.interface.get_blocks.return_value = async_coro({'babe': {'block_hash': 'babe', 'block_bytes': b'raw'}})
        self.repo.headers.get_headers_since_height.return_value = [{'block_hash': 'babe', 'block_height': 10}]
        self.repo.blockchain.get_block.return_value = None
        self.repo.blockchain.save_blocks.side_effect = ValueError

        self.loop.run_until_complete(self.sut.check())
        Mock.assert_called_once_with(
            self.repo.blockchain.save_blocks, {'block_hash': 'babe', 'block_bytes': b'raw'}
        )
        Mock.assert_called_once_with(self.repo.headers.get_best_header)
        Mock.assert_called_once_with(self.interface.get_blocks, 'babe')
        Mock.assert_called_once_with(self.repo.headers.get_headers_since_height, 9, limit=10)
        Mock.assert_called_once_with(self.repo.blockchain.get_block, 'babe', with_transactions=False)
        self.assertEqual(self.sut._last_processed_block, {'block_hash': 'cafe', 'block_height': 9})

    def test_check_blockchain_local_a_lot_behind(self):
        """
        something bad happened around block 16, we saved it, we didn't tracked it and we start over 5 blocks later.
        also, the last block tracked by the blockheader reactor was stuck ad block 9.
        basically, everything is screwed up, but we recover and download the needed blocks.
        """
        self.sut.set_last_processed_block({'block_hash': 'cafe', 'block_height': 9})
        self.repo.headers.get_best_header.return_value = {'block_hash': 'babe', 'block_height': 20}
        self.interface.get_blocks.return_value = async_coro(
            {
                'block17': {'block_hash': 'block17', 'block_bytes': b'raw'},
                'block18': {'block_hash': 'block18', 'block_bytes': b'raw'},
                'block19': {'block_hash': 'block19', 'block_bytes': b'raw'},
                'block20': {'block_hash': 'block20', 'block_bytes': b'raw'}
            }
        )
        self.repo.headers.get_headers_since_height.return_value = [
            {'block_hash': 'block16', 'block_height': 16},
            {'block_hash': 'block17', 'block_height': 17},
            {'block_hash': 'block18', 'block_height': 18},
            {'block_hash': 'block19', 'block_height': 19},
            {'block_hash': 'block20', 'block_height': 20}
        ]
        self.repo.blockchain.get_block.side_effect = [
            {'block_hash': 'block16', 'block_bytes': b'raw'}, None, None, None, None
        ]
        self.repo.blockchain.save_blocks.side_effect = [
            {'block_hash': 'block17', 'block_bytes': b'raw'},
            {'block_hash': 'block18', 'block_bytes': b'raw'},
            {'block_hash': 'block19', 'block_bytes': b'raw'},
            {'block_hash': 'block20', 'block_bytes': b'raw'}
        ]
        self.repo.blockchain.save_blocks.side_effect = lambda *x: x

        self.loop.run_until_complete(self.sut.check())

        Mock.assert_called_once_with(self.repo.headers.get_best_header)
        Mock.assert_called_once_with(self.repo.headers.get_headers_since_height, 15, limit=10)
        Mock.assert_has_calls(
            self.repo.blockchain.get_block,
            calls=[
                call('block16', with_transactions=False),
                call('block17', with_transactions=False),
                call('block18', with_transactions=False),
                call('block19', with_transactions=False),
                call('block20', with_transactions=False)
            ]
        )
        Mock.assert_called_once_with(self.interface.get_blocks, 'block17', 'block18', 'block19', 'block20')

        Mock.assert_called_once_with(
            self.repo.blockchain.save_blocks,
            {'block_hash': 'block17', 'block_bytes': b'raw'},
            {'block_hash': 'block18', 'block_bytes': b'raw'},
            {'block_hash': 'block19', 'block_bytes': b'raw'},
            {'block_hash': 'block20', 'block_bytes': b'raw'}
        )
        self.assertEqual(self.sut._last_processed_block, {'block_hash': 'block20', 'block_height': 20})

    def test_check_corners_orphaned(self):
        self.sut.set_last_processed_block({'block_hash': 'cafe', 'block_height': 9})
        self.assertEqual(self.sut._last_processed_block, {'block_hash': 'cafe', 'block_height': 9})
        self.repo.headers.get_best_header.return_value = {'block_hash': 'babe', 'block_height': 8}
        self.loop.run_until_complete(self.sut.check())
        self.assertEqual(self.sut._last_processed_block, None)

    def test_check_corners_reorg(self):
        self.sut.set_last_processed_block({'block_hash': 'cafe', 'block_height': 9})
        self.assertEqual(self.sut._last_processed_block, {'block_hash': 'cafe', 'block_height': 9})
        self.repo.headers.get_best_header.return_value = {'block_hash': 'babe', 'block_height': 9}
        self.loop.run_until_complete(self.sut.check())
        self.assertEqual(self.sut._last_processed_block, None)

    def test_bootstrap(self):
        self.interface.pool = Mock(_busy_peers=[])
        self.interface.pool.required_connections = 1
        self.interface.pool.established_connections = []

        async def add_peers():
            await asyncio.sleep(0)
            for _ in range(0, 8):
                self.interface.pool.established_connections.append(1)

        self.repo.headers.get_best_header.return_value = {'block_hash': 'cafe', 'block_height': 6}
        self.repo.headers.get_headers_since_height.return_value = [
            {'block_hash': 'block2', 'block_height': 2},
            {'block_hash': 'block3', 'block_height': 3},
            {'block_hash': 'block4', 'block_height': 4},
            {'block_hash': 'block5', 'block_height': 5},
            {'block_hash': 'block6', 'block_height': 6}
        ]
        self.repo.blockchain.get_block.side_effect = [
            {'block_hash': 'block2', 'block_bytes': b'raw'}, None, None, None, None
        ]
        self.interface.get_block.side_effect = [
            async_coro({'block_hash': 'block3', 'block_bytes': b'raw'}),
            async_coro({'block_hash': 'block4', 'block_bytes': b'raw'}),
            async_coro({'block_hash': 'block5', 'block_bytes': b'raw'}),
            async_coro({'block_hash': 'block6', 'block_bytes': b'raw'}),
        ]

        self.loop.run_until_complete(
            asyncio.gather(
                add_peers(),
                self.sut.bootstrap_blocks()
            )
        )
        Mock.assert_has_calls(
            self.repo.blockchain.save_block,
            calls=[
                call({'block_hash': 'block3', 'block_bytes': b'raw'}),
                call({'block_hash': 'block4', 'block_bytes': b'raw'}),
                call({'block_hash': 'block5', 'block_bytes': b'raw'}),
                call({'block_hash': 'block6', 'block_bytes': b'raw'})
            ]
        )



from typing import List, Dict
from sqlalchemy.exc import IntegrityError
from spruned.application.abstracts import HeadersRepository
from spruned.application.logging_factory import Logger
from spruned.daemon import exceptions
from spruned.application import database


class HeadersSQLiteRepository(HeadersRepository):
    def __init__(self, session):
        self.session = session
        self._cache = None

    def set_cache(self, cache):
        self._cache = cache

    @staticmethod
    def _header_model_to_dict(header: database.Header, nextblockhash: (None, str), prevblockhash: (None, str)) -> Dict:
        res = {
            'block_height': header.blockheight,
            'block_hash': header.blockhash,
            'header_bytes': header.data,
            'next_block_hash': nextblockhash,
            'prev_block_hash': prevblockhash
        }
        if not res['prev_block_hash']:
            res.pop('prev_block_hash')
        if not res['next_block_hash']:
            res.pop('next_block_hash')
        return res

    def get_best_header(self):
        session = self.session()
        res = session.query(database.Header).order_by(database.Header.blockheight.desc()).limit(1).one_or_none()
        if not res:
            return
        prevblockhash = res.blockheight != 0 and self.get_block_hash(res.blockheight - 1)
        res = self._header_model_to_dict(res, None, prevblockhash)
        return res

    def get_header_at_height(self, height: int):
        blockhash = self.get_block_hash(height)
        return self.get_block_header(blockhash)

    def get_headers_since_height(self, height: int, limit=None):
        session = self.session()
        query = session.query(database.Header).filter(database.Header.blockheight >= height)\
            .order_by(database.Header.blockheight.asc())
        if limit is not None:
            query = query.limit(limit)
        headers = query.all()
        return headers and [
            self._header_model_to_dict(
                h,
                nextblockhash=self.get_block_hash(h.blockheight+1),
                prevblockhash=h.blockheight != 0 and self.get_block_hash(h.blockheight-1)
            ) for h in headers
        ] or []

    def get_headers(self, *blockhashes: str):
        session = self.session()
        headers = session.query(database.Header).filter(database.Header.blockhash.in_(blockhashes))\
            .order_by(database.Header.blockheight.asc()).all()
        if set([h.blockhash for h in headers]) - set(blockhashes):
            # not sure if all raises, investigate # FIXME
            raise exceptions.HeadersInconsistencyException
        return headers and [
            self._header_model_to_dict(
                h,
                nextblockhash=self.get_block_hash(h.blockheight+1),
                prevblockhash=h.blockheight != 0 and self.get_block_hash(h.blockheight-1)
            ) for h in headers
        ] or []

    @database.atomic
    def save_header(self, blockhash: str, blockheight: int, headerbytes: bytes, prev_block_hash: str):
        session = self.session()

        def _save():
            model = database.Header(blockhash=blockhash, blockheight=blockheight, data=headerbytes)
            session.add(model)
            try:
                session.flush()
            except IntegrityError:
                raise exceptions.HeadersInconsistencyException
            return model

        if blockheight == 0:
            model = _save()
            prev_block = None
        else:
            prev_block = session.query(database.Header).filter_by(blockheight=blockheight - 1).one()
            if prev_block.blockhash != prev_block_hash:
                raise exceptions.HeadersInconsistencyException

            model = _save()
        return model and self._header_model_to_dict(
            model, prevblockhash=prev_block and prev_block.blockhash, nextblockhash=None
        )

    @database.atomic
    def save_headers(self, headers: List[Dict]):
        session = self.session()
        for i, header in enumerate(headers):
            if i == 0 and header['block_height'] != 0:
                prev_block = session.query(database.Header).filter_by(blockheight=header['block_height'] - 1).one()
                if prev_block.blockhash != header['prev_block_hash']:
                    Logger.repository.exception('Integrity Error on check prev block hash')
                    raise exceptions.HeadersInconsistencyException
            model = database.Header(
                blockhash=header['block_hash'],
                blockheight=header['block_height'],
                data=header['header_bytes']
            )
            session.add(model)
        try:
            session.flush()
        except (IntegrityError, AssertionError) as e:
            Logger.repository.exception('Integrity Error on save_headers')
            raise exceptions.HeadersInconsistencyException
        return headers

    @database.atomic
    def remove_headers_after_height(self, blockheight: int):
        session = self.session()
        headers = session.query(database.Header).filter(database.Header.blockheight >= blockheight).all()
        for header in headers:
            session.delete(header)
        session.flush()

    @database.atomic
    def remove_header_at_height(self, blockheight: int) -> Dict:
        session = self.session()
        header = session.query(database.Header).filter(database.Header.blockheight == blockheight).one()
        removing_dict = self._header_model_to_dict(header, "", "")
        session.delete(header)
        session.flush()
        return removing_dict

    def get_block_hash(self, blockheight: int):
        session = self.session()
        header = session.query(database.Header).filter_by(blockheight=blockheight).one_or_none()
        return header and header.blockhash

    def get_block_height(self, blockhash: str):
        session = self.session()
        header = session.query(database.Header).filter_by(blockhash=blockhash).one_or_none()
        return header and header.blockheight

    def get_block_header(self, blockhash: str):
        session = self.session()
        header = session.query(database.Header).filter_by(blockhash=blockhash).one_or_none()
        if not header:
            return
        nextblockhash = self.get_block_hash(header.blockheight + 1)
        prevblockhash = header.blockheight != 0 and self.get_block_hash(header.blockheight - 1)
        return self._header_model_to_dict(header, nextblockhash, prevblockhash)



import asyncio
import pickle
import time

from unittest import TestCase
from unittest.mock import Mock, create_autospec

from spruned.application.cache import CacheAgent
from spruned.repositories.repository import Repository


class TestCacheAgent(TestCase):
    def setUp(self):
        self.loop = asyncio.get_event_loop()
        self.session = Mock()
        self.repository = create_autospec(Repository)
        self.repository.ldb = self.session
        self.repository.blockchain.set_cache.return_value = None
        self.loopmock = Mock()
        self.delayer = Mock
        self.sut = CacheAgent(
            self.repository,
            1024,
            self.loopmock,
            self.delayer
        )
        self.basedata = {
            'keys': {
                'cafe': {
                    'saved_at': 123,
                    'size': 16,
                    'key': 'cafe'
                },
                'babe': {
                    'saved_at': 124,
                    'size': 32,
                    'key': 'babe'
                }
            },
            'total': 48
        }

    def test_init_no_data(self):
        self.session.get.return_value = None
        self.sut.init()

    def test_init_data(self):
        dump = pickle.dumps([['cafe', 123, 16], ['babe', 124, 32]])
        self.session.get.return_value = dump
        self.sut.init()
        self.assertEqual(self.sut._last_dump_size, 48)
        self.assertEqual(self.sut.index, self.basedata)
        self.assertEqual(self.sut.index['total'], 48)

    def test_save_data(self):
        dump = pickle.dumps([['cafe', 123, 16], ['babe', 124, 32]])
        self.session.get.return_value = dump
        self.sut.init()
        self.sut.dump()
        Mock.assert_called_once_with(self.session.put, b'cache_index', dump)

    def test_track(self):
        dump = pickle.dumps([['cafe', 123, 16], ['babe', 124, 32]])
        self.session.get.return_value = dump
        self.sut.init()
        now = int(time.time())
        self.sut.track('ffff', 64)
        self.assertEqual(self.sut.index['total'], 112)
        self.assertEqual(self.sut.index['keys']['ffff']['key'], 'ffff')
        self.assertEqual(self.sut.index['keys']['ffff']['size'], 64)
        self.assertTrue(now - 1 <= self.sut.index['keys']['ffff']['saved_at'] <= now + 1)

    def test_track_empty_index(self):
        now = int(time.time())
        self.sut.track('ffff', 64)
        self.assertEqual(self.sut.index['total'], 64)
        self.assertEqual(self.sut.index['keys']['ffff']['key'], 'ffff')
        self.assertEqual(self.sut.index['keys']['ffff']['size'], 64)
        self.assertTrue(now - 1 <= self.sut.index['keys']['ffff']['saved_at'] <= now + 1)

    def test_check_ok(self):
        dump = pickle.dumps([['cafe', 123, 16], ['babe', 124, 32]])
        self.session.get.return_value = dump
        self.sut.init()
        self.loop.run_until_complete(self.sut.check())
        Mock.assert_not_called(self.session.put)

    def test_check_new_data_saved(self):
        ldump = [['cafe', 123, 16], ['babe', 124, 32]]
        self.session.get.side_effect = [pickle.dumps(ldump), True, True, True]
        self.sut.init()
        self.sut.track('ffff', 64)
        ldump.append(['ffff', self.sut.index['keys']['ffff']['saved_at'], 64])
        self.loop.run_until_complete(self.sut.check())
        Mock.assert_called_once_with(self.session.put, b'cache_index', pickle.dumps(ldump))

    def test_check_new_data_saved_size_limit_exceeded(self):
        """
        this is the real naming convention for indexes
        :return:
        """
        ldump = [[b'\x01.cafe', 123, 512], [b'\x00.bbbb', 123, 128], [b'\x01.babe', 124, 128]]
        self.session.get.side_effect = [pickle.dumps(ldump), True, True, True, True]
        self.sut.init()
        self.sut.track(b'\x01.ffff', 640)
        ldump.append([b'\x01.ffff', self.sut.index['keys'][b'\x01.ffff']['saved_at'], 640])
        ldump = ldump[1:]
        self.loop.run_until_complete(self.sut.check())
        Mock.assert_called_once_with(self.repository.blockchain.remove_block, b'cafe')
        Mock.assert_called_once_with(self.session.put, b'cache_index', pickle.dumps(ldump))



import async_timeout
from pycoinnet.dnsbootstrap import dns_bootstrap_host_port_q
from pycoinnet.networks import MAINNET, TESTNET
import asyncio

from spruned.application.logging_factory import Logger


async def dns_bootstrap_servers(network=TESTNET, howmany=50):  # pragma: no cover
    host_q = dns_bootstrap_host_port_q(network)
    ad = []
    while 1:
        item = host_q.get()
        try:
            async with async_timeout.timeout(1):
                peer = await item
        except asyncio.TimeoutError:
            try:
                item.close()
            except asyncio.CancelledError:
                Logger.p2p.debug('Cancelled')
            break
        ad.append(peer)
    return ad





import asyncio
import unittest
from unittest.mock import Mock, create_autospec, call
import time
from spruned.application.abstracts import HeadersRepository
from spruned.application.context import ctx
from spruned.daemon import exceptions
from spruned.daemon.electrod.electrod_interface import ElectrodInterface
from spruned.daemon.tasks.headers_reactor import HeadersReactor
from test.utils import async_coro, coro_call, in_range, make_headers
import warnings


class TestHeadersReactor(unittest.TestCase):
    def setUp(self):
        self.repo = create_autospec(HeadersRepository)
        self.interface = create_autospec(ElectrodInterface)
        self.electrod_loop = Mock()
        self.electrod_loop.create_task.side_effect = lambda x: x
        self.delay_task_runner = Mock()
        self.sut = HeadersReactor(
            self.repo, self.interface, loop=self.electrod_loop, delayed_task=self.delay_task_runner
        )
        self.loop = asyncio.get_event_loop()
        warnings.filterwarnings("ignore")

    def tearDown(self):
        self.repo.reset_mock()
        self.interface.reset_mock()
        self.electrod_loop.reset_mock()
        self.delay_task_runner.reset_mock()

    def test_last_header_received_less_than_min_polling_interval(self):
        """
        test reactor.check_headers method
        last header is saved less than polling interval value (default: 660)
        """
        header_timestamp = int(time.time()) - 100
        loc_header = {"block_height": 1, "block_hash": "ff"*32, "timestamp": header_timestamp}

        self.sut.synced = True
        self.sut.set_last_processed_header(loc_header)
        self.loop.run_until_complete(self.sut.check_headers())
        self.assertFalse(self.sut.lock.locked())

        Mock.assert_called_once_with(self.delay_task_runner, coro_call('check_headers'), in_range(559, 560))
        self.assertEqual(0, len(self.interface.method_calls))
        self.assertEqual(1, len(self.electrod_loop.method_calls))
        self.assertEqual(0, len(self.repo.method_calls))

    def test_no_new_network_best_header(self):
        """
        test reactor.check_headers method
        interface didn't returned any best header. will try again in
        """
        header_timestamp = int(time.time()) - self.sut.new_headers_fallback_poll_interval - 1
        loc_header = {"block_height": 1, "block_hash": "ff" * 32, "timestamp": header_timestamp}
        self.interface.get_header.return_value = async_coro(None)
        self.sut.synced = True
        self.sut.set_last_processed_header(loc_header)
        self.loop.run_until_complete(self.sut.check_headers())
        self.assertFalse(self.sut.lock.locked())

        Mock.assert_called_once_with(self.delay_task_runner, coro_call('check_headers'), 660)
        Mock.assert_called_with(self.interface.get_header, 2, fail_silent_out_of_range=True, get_peer=True)
        Mock.assert_called_once_with(self.electrod_loop.create_task, self.delay_task_runner())

        self.assertEqual(1, len(self.interface.method_calls))
        self.assertEqual(1, len(self.electrod_loop.method_calls))
        self.assertEqual(0, len(self.repo.method_calls))

    def test_fallback_check_error(self):
        """
        test reactor.check_headers method
        interface didn't returned any best header. will try again in
        """
        header_timestamp = int(time.time()) - self.sut.new_headers_fallback_poll_interval - 1
        loc_header = {"block_height": 1, "block_hash": "ff" * 32, "timestamp": header_timestamp}
        self.interface.get_header.side_effect = [exceptions.NoPeersException]
        self.sut.synced = True
        self.sut.set_last_processed_header(loc_header)
        self.loop.run_until_complete(self.sut.check_headers())
        self.assertFalse(self.sut.lock.locked())

        Mock.assert_called_once_with(self.delay_task_runner, coro_call('check_headers'), 30)
        Mock.assert_called_with(self.interface.get_header, 2, fail_silent_out_of_range=True, get_peer=True)
        Mock.assert_called_once_with(self.electrod_loop.create_task, self.delay_task_runner())

        self.assertEqual(1, len(self.interface.method_calls))
        self.assertEqual(1, len(self.electrod_loop.method_calls))
        self.assertEqual(0, len(self.repo.method_calls))

    def test_error_fetching_new_network_best_header(self):
        """
        test reactor.check_headers method
        interface can't find, due race conditions, peers to fetch the headers
        NoPeersException is raised
        """
        header_timestamp = int(time.time()) - self.sut.new_headers_fallback_poll_interval - 1
        loc_header = {"block_height": 1, "block_hash": "ff" * 32, "timestamp": header_timestamp}
        self.interface.get_header.side_effect = exceptions.NoPeersException
        self.sut.synced = True
        self.sut.set_last_processed_header(loc_header)
        self.loop.run_until_complete(self.sut.check_headers())
        self.assertFalse(self.sut.lock.locked())

        Mock.assert_called_once_with(self.delay_task_runner, coro_call('check_headers'), 30)
        Mock.assert_called_with(self.interface.get_header, 2, fail_silent_out_of_range=True, get_peer=True)
        Mock.assert_called_once_with(self.electrod_loop.create_task, self.delay_task_runner())

        self.assertEqual(1, len(self.interface.method_calls))
        self.assertEqual(1, len(self.electrod_loop.method_calls))
        self.assertEqual(0, len(self.repo.method_calls))

    def test_network_header_behind(self):
        """
        test reactor.on_header method

        network header behind: remote peer outdated, skip
        """
        self.sut.synced = True
        now = int(time.time())
        net_header = {
            'block_height': 1, 'block_hash': 'aa' * 32,
            'timestamp': now - self.sut.new_headers_fallback_poll_interval + 10
        }
        loc_header = {
            'block_height': 2, 'block_hash': 'ff'*32, 'timestamp': now-self.sut.new_headers_fallback_poll_interval+10
        }
        peer = Mock(server_info='mock_peer')
        peer.close.return_value = True
        self.repo.get_best_header.return_value = loc_header
        self.repo.get_header_at_height.return_value = net_header
        self.interface.disconnect_from_peer.return_value = async_coro(True)
        self.loop.run_until_complete(self.sut.on_new_header(peer, net_header))

        Mock.assert_called_once_with(self.repo.get_header_at_height, 1)
        self.assertEqual(1, len(self.interface.method_calls))
        self.assertEqual(0, len(self.electrod_loop.method_calls))
        self.assertEqual(2, len(self.repo.method_calls))

    def test_network_equal(self):
        """
        test reactor.on_header

        electrod aligned with remote peers, ensure consistency of received header with local db.
        """
        self.sut.synced = True
        now = int(time.time())
        peer = Mock(server_info='mock_peer')
        net_header = loc_header = {'block_height': 2, 'block_hash': 'ff' * 32, 'timestamp': now-1000}
        self.sut.set_last_processed_header(loc_header)
        self.loop.run_until_complete(self.sut.on_new_header(peer, net_header))
        self.assertEqual(0, len(self.interface.method_calls))
        self.assertEqual(0, len(self.electrod_loop.method_calls))
        self.assertEqual(0, len(self.repo.method_calls))

    def test_received_new_single_header(self):
        """
        test reactor.on_header method

        header ff*32 is saved at height 2020
        received a new header at height 2021 with hash aa*32
        header is saved to disk
        best header is updated at height 2021
        """
        self.sut.synced = False
        peer = Mock(server_info='mock_peer')
        loc_header = {'block_height': 2020, 'block_hash': 'ff' * 32}
        net_header = {'block_height': 2021, 'block_hash': 'aa' * 32, 'prev_block_hash': 'ff'*32, 'header_bytes': b''}
        self.interface.get_header.return_value = async_coro(net_header)

        self.repo.get_best_header.return_value = loc_header

        self.loop.run_until_complete(self.sut.on_new_header(peer, net_header))
        Mock.assert_called_once_with(self.repo.get_best_header)
        Mock.assert_called_once_with(
            self.repo.save_header,
            net_header['block_hash'], net_header['block_height'],
            net_header['header_bytes'], net_header['prev_block_hash']
        )
        self.assertEqual(0, len(self.interface.method_calls))
        self.assertEqual(0, len(self.electrod_loop.method_calls))
        self.assertEqual(2, len(self.repo.method_calls))

    def test_remove_orphan_header_previously_saved(self):
        """
        test reactor.on_header

        header ff*32 is saved at height 2020
        received a new header at height 2020 with hash 00*32
        oops. loop is paused for 20 seconds to allows the network to reorg.
        header at height 2020 is requested to the interface, header 00*32 is returned
        reactor compare headers
        ff*32 is deleted by the repository
        orphaned blockheader is saved somewhere for future usage
        """
        header_timestamp = int(time.time())
        loc_header = {"block_height": 2020, "block_hash": "ff"*32, "timestamp": header_timestamp - 10}
        net_header = {"block_height": 2020, "block_hash": "00"*32, "timestamp": header_timestamp}
        self.sut.synced = True
        self.sut.sleep_time_on_inconsistency = 1
        self.sut.set_last_processed_header(loc_header)

        peer = Mock(server_info='mock_peer')
        self.interface.handle_peer_error.return_value = async_coro(True)
        self.repo.get_block_hash.return_value = 'ff' * 32
        self.interface.get_header.return_value = async_coro(net_header)
        self.repo.remove_header_at_height.return_value = loc_header
        self.repo.get_best_header.return_value = loc_header
        self.loop.run_until_complete(self.sut.on_new_header(peer, net_header))

        Mock.assert_called_with(self.repo.remove_header_at_height, 2020)
        Mock.assert_not_called(peer.close)

        self.assertEqual(2, len(self.interface.method_calls))
        self.assertEqual(3, len(self.repo.method_calls))
        self.assertEqual(0, len(self.electrod_loop.method_calls))
        self.assertFalse(self.sut.synced)
        self.assertIn(loc_header, self.sut.orphans_headers)

    def test_orphan_received_saved_ok(self):
        """
        test reactor.on_header

        received an orphan block, but local is ok, other peers says
        """
        header_timestamp = int(time.time())
        loc_header = {"block_height": 2020, "block_hash": "ff" * 32, "timestamp": header_timestamp - 10}
        net_header = {"block_height": 2020, "block_hash": "00" * 32, "timestamp": header_timestamp}
        self.sut.synced = True
        self.sut.sleep_time_on_inconsistency = 1
        self.sut.set_last_processed_header(loc_header)

        peer = Mock(server_info='mock_peer')
        self.interface.get_header.return_value = async_coro(loc_header)
        self.interface.handle_peer_error.side_effect = [async_coro(True), async_coro(True)]

        self.repo.get_best_header.return_value = loc_header
        self.repo.get_block_hash.return_value = 'ff' * 32
        self.repo.remove_header_at_height.return_value = loc_header
        self.loop.run_until_complete(self.sut.on_new_header(peer, net_header))

        Mock.assert_not_called(self.repo.remove_header_at_height)
        Mock.assert_called_with(self.interface.handle_peer_error, peer)
        Mock.assert_called_with(self.interface.get_header, 2020, fail_silent_out_of_range=True)

        self.assertEqual(3, len(self.interface.method_calls))
        self.assertEqual(2, len(self.repo.method_calls))
        self.assertEqual(0, len(self.electrod_loop.method_calls))
        self.assertTrue(self.sut.synced)
        self.assertNotIn(loc_header, self.sut.orphans_headers)

    def test_local_db_behind_100_headers(self):
        """
        test reactor.on_new_header

        header ff*32 is saved at height 2020
        received a new header at height 2120 with hash aa*32
        chunk from header 2016 is requested
        headers from 2021 to 2120 are saved in the db
        best header is updated at height 2120
        :return:
        """
        header_timestamp = int(time.time())
        peer = Mock(server_info='mock_peer')
        loc_header = {"block_height": 2020, "block_hash": "ff" * 32, "timestamp": header_timestamp - 6000}
        net_header = {"block_height": 2120, "block_hash": "aa" * 32, "timestamp": header_timestamp}
        self.sut.synced = True
        self.sut.set_last_processed_header(loc_header)
        self.repo.get_best_header.return_value = loc_header
        _headers = make_headers(2017, 2120, '00'*32)
        self.interface.get_headers_in_range_from_chunks.side_effect = [async_coro((Mock(), _headers)), async_coro(None)]
        self.interface.get_header.return_value = async_coro(net_header)
        self.repo.save_headers.side_effect = lambda x, **k: x

        self.loop.run_until_complete(self.sut.on_new_header(peer, net_header))

        Mock.assert_called_with(self.repo.save_headers, [h for h in _headers if h['block_height'] > 2020])
        Mock.assert_not_called(peer.close)
        self.assertEqual(self.sut._last_processed_header, _headers[-1])
        self.assertEqual(self.sut.synced, True)
        self.assertEqual(1, len(self.interface.method_calls))
        self.assertEqual(2, len(self.repo.method_calls))
        self.assertEqual(0, len(self.electrod_loop.method_calls))

    def test_received_header_that_doesnt_link_to_previous(self):
        """
        on_new_header

        header ff*32 is saved at height 2020
        header aa*32 at height 2021 have prev_block_hash 00*32
        header 2021 is fetched and verified from the network
        header 2020 with hash 00*32 is fetched from 3 peers in the network
        chunk from header 2016 is requested
        headers from 2016 are deleted from the database
        headers from 2016 to 2021 are saved in the db
        best header is updated at height 2021 with hash aa*32
        blockhash ff*32 is saved as orphan and will be ignored to avoid others peers to cause recursive inconsistencies
        """
        header_timestamp = int(time.time())
        peer = Mock(server_info='mock_peer')
        loc_header = {"block_height": 2020, "block_hash": "ff" * 32, "timestamp": header_timestamp - 1000}
        net_header = {
            "block_height": 2021,
            "block_hash": "aa" * 32,
            "timestamp": header_timestamp,
            "prev_block_hash": "00"*32,
            "header_bytes": b""
        }
        self.sut.synced = True
        self.sut.set_last_processed_header(loc_header)
        self.interface.get_header.return_value = async_coro(net_header)
        self.repo.get_best_header.return_value = loc_header

        self.repo.save_header.side_effect = [exceptions.HeadersInconsistencyException]
        self.loop.run_until_complete(self.sut.on_new_header(peer, net_header))
        Mock.assert_called_once_with(self.repo.remove_headers_after_height, 2016)
        self.assertEqual(self.repo.get_best_header.call_count, 2)
        Mock.assert_called_once_with(self.repo.remove_headers_after_height, 2016)

        self.assertIsNone(self.sut._last_processed_header)
        self.assertEqual(0, len(self.interface.method_calls), msg=str(self.interface.method_calls))
        self.assertEqual(4, len(self.repo.method_calls), msg=str(self.repo.method_calls))
        self.assertEqual(0, len(self.electrod_loop.method_calls), msg=str(self.electrod_loop.method_calls))
        self.assertFalse(self.sut.synced)
        self.assertFalse(self.sut.lock.locked())

    def test_no_local_headers(self):
        """
        a new header is received, best_height is 3000
        there is no local header
        the chunk 0 is requested, but there are no peers available due race condition
        then the chunk 0 is fetched and saved
        then the chunk 1 is fetched and saved
        new height is 3000
        reactor is in sync with the network
        """
        header_timestamp = int(time.time())
        net_header = {
            "block_height": 3000,
            "block_hash": "cc" * 32,
            "timestamp": header_timestamp,
            "prev_block_hash": "00" * 32,
            "header_bytes": b"0"*80
        }
        peer = Mock(server_info='mock_peer')
        loc_header = None
        self.repo.get_best_header.return_value = loc_header
        _chunk_1 = make_headers(0, 2015, ctx.get_network()['checkpoints'][0])
        _chunk_2 = make_headers(2016, 2999, _chunk_1[-1]['block_height'])
        _chunk_2.append(net_header)

        self.interface.get_headers_in_range_from_chunks.side_effect = [
            exceptions.NoPeersException,
            async_coro(None),
            async_coro((Mock(), _chunk_1)),
            async_coro((Mock(), _chunk_2))
        ]
        self.interface.get_header.return_value = async_coro((Mock(), net_header))
        self.repo.save_headers.side_effect = lambda x, **k: x

        self.loop.run_until_complete(self.sut.on_new_header(peer, net_header))
        self.assertEqual(self.sut._last_processed_header, net_header)
        self.assertTrue(self.sut.synced)
        self.assertEqual(4, len(self.interface.method_calls))
        self.assertEqual(5, len(self.repo.method_calls))
        self.assertEqual(0, len(self.electrod_loop.method_calls))

    def test_on_new_headers_5_blocks_behind(self):
        """
        local height is 10
        remote height is 15
        5 headers are requested with interface.get_headers_in_range api (no chunks)
        due race conditions, two errors are triggered: no peers, empty response
        the third retry headers are fetched
        headers are saved
        best height is 15
        sync true
        """
        header_timestamp = int(time.time())
        net_header = {
            "block_height": 15,
            "block_hash": "55" * 32,
            "timestamp": header_timestamp,
            "prev_block_hash": "54" * 32,
            "header_bytes": b"0"*80
        }
        peer = Mock(server_info='mock_peer')
        loc_header = {
            "block_height": 10,
            "block_hash": "bb" * 32,
            "timestamp": header_timestamp - 600*5,
            "prev_block_hash": "aa" * 32,
            "header_bytes": b"0"*80
        }
        self.repo.get_best_header.return_value = loc_header
        headers = make_headers(11, 15, 'bb'*32) + [net_header]

        self.interface.get_headers_in_range.side_effect = [
            exceptions.NoPeersException,
            async_coro(None),
            async_coro(headers),
        ]
        self.interface.get_header.return_value = async_coro(net_header)
        self.repo.save_headers.side_effect = lambda x, **k: x

        self.loop.run_until_complete(self.sut.on_new_header(peer, net_header))
        self.assertEqual(self.sut._last_processed_header, net_header)
        self.assertTrue(self.sut.synced)
        self.assertEqual(3, len(self.interface.method_calls))
        self.assertEqual(4, len(self.repo.method_calls))
        self.assertEqual(0, len(self.electrod_loop.method_calls))

    def test_check_headers_new_header(self):
        """
        test on fallback method check_headers
        new header received
        """
        header_timestamp = int(time.time()) - 100
        loc_header = {
            "block_height": 1,
            "block_hash": "ff"*32, "timestamp": header_timestamp - 700, 'header_bytes': b'', 'prev_block_hash': '00'*32
        }
        net_header = {
            "block_height": 2,
            "block_hash": "aa"*32, "timestamp": header_timestamp, 'header_bytes': b'', 'prev_block_hash': 'ff'*32
        }
        peer = Mock()
        self.interface.get_header.side_effect = [async_coro((peer, net_header))]
        self.sut.synced = True
        self.sut.set_last_processed_header(loc_header)
        self.assertFalse(self.sut.lock.locked())
        peer = Mock(server_info='mock_peer')
        peer.close.return_value = True
        self.repo.get_best_header.return_value = loc_header
        self.repo.save_header.side_effect = lambda a, b, c, d: True

        self.loop.run_until_complete(self.sut.check_headers())
        Mock.assert_called_once_with(self.delay_task_runner, coro_call('check_headers'), 30)
        self.assertEqual(1, len(self.interface.method_calls))
        self.assertEqual(2, len(self.electrod_loop.method_calls))
        self.assertEqual(0, len(self.repo.method_calls))
        self.assertTrue(self.sut.synced)

    def test_on_new_headers_new_header(self):
        """
        test on fallback method check_headers
        new header received
        """
        header_timestamp = int(time.time()) - 100
        loc_header = {
            "block_height": 1,
            "block_hash": "ff"*32, "timestamp": header_timestamp - 700, 'header_bytes': b'', 'prev_block_hash': '00'*32
        }
        net_header = {
            "block_height": 2,
            "block_hash": "aa"*32, "timestamp": header_timestamp, 'header_bytes': b'', 'prev_block_hash': 'ff'*32
        }
        self.interface.get_header.side_effect = [async_coro(net_header)]
        self.sut.synced = True
        self.sut.set_last_processed_header(loc_header)
        self.assertFalse(self.sut.lock.locked())
        peer = Mock(server_info='mock_peer')
        self.repo.get_best_header.return_value = loc_header
        self.repo.save_header.side_effect = lambda a, b, c, d: True

        self.loop.run_until_complete(self.sut.on_new_header(peer, net_header))
        self.assertEqual(self.repo.get_best_header.call_count, 1)
        Mock.assert_called_once_with(
            self.repo.save_header,
            net_header['block_hash'], net_header['block_height'],
            net_header['header_bytes'], net_header['prev_block_hash']
        )

    def test_check_new_header_same_header(self):
        """
        test on fallback method check_headers
        no new header received
        """
        header_timestamp = int(time.time()) - 100
        loc_header = net_header = {
            "block_height": 1,
            "block_hash": "ff"*32, "timestamp": header_timestamp - 700, 'header_bytes': b'', 'prev_block_hash': '00'*32
        }
        peer = Mock(server_info='mock_peer')
        self.interface.get_header.side_effect = [async_coro((peer, net_header))]
        self.sut.synced = True
        self.sut.set_last_processed_header(loc_header)
        self.assertFalse(self.sut.lock.locked())
        self.loop.run_until_complete(self.sut.check_headers())
        Mock.assert_called_once_with(self.delay_task_runner, coro_call('check_headers'), in_range(655, 660))
        Mock.assert_has_calls(
            self.interface.get_header,
            calls=[
                call(2, fail_silent_out_of_range=True, get_peer=True),
            ], any_order=True
        )
        self.assertEqual(1, len(self.interface.method_calls))
        self.assertEqual(1, len(self.electrod_loop.method_calls))
        self.assertEqual(0, len(self.repo.method_calls))

    def test_mixed(self):
        self.sut._sync_errors = 101
        with self.assertRaises(exceptions.SprunedException):
            self.loop.run_until_complete(self.sut.check_headers())
        self.sut._sync_errors = 0

        # I'm not sure this can happen. Anyway, just drop the call
        self.loop.run_until_complete(self.sut.on_new_header(Mock(), {}))

        self.loop.run_until_complete(self.sut.start())
        self.assertEqual(self.electrod_loop.create_task.call_count, 1)
        self.assertEqual(len(self.electrod_loop.method_calls), 1)

    def test_no_store_headers(self):
        self.sut.store_headers = False
        self.loop.run_until_complete(self.sut.on_connected())
        Mock.assert_not_called(self.electrod_loop.create_task)

    def test_not_synced_yet_fallback_task(self):
        self.sut.synced = False
        self.loop.run_until_complete(self.sut.check_headers())
        Mock.assert_called_once_with(self.delay_task_runner, coro_call('check_headers'), in_range(29, 30))

    def test_locked_get_header_fallback_task(self):
        self.sut.synced = True
        self.loop.run_until_complete(self.sut.lock.acquire())
        self.loop.run_until_complete(self.sut.check_headers())
        self.sut.lock.release()
        Mock.assert_called_once_with(self.delay_task_runner, coro_call('check_headers'), in_range(29, 30))

    def test_on_new_header_inconsistency_between_peers(self):
        """
        test reactor.on_header method

        header ff*32 is saved at height 2020
        received a new header at height 2021 with hash aa*32
        header is saved to disk
        best header is updated at height 2021
        """
        self.sut.synced = False
        peer = Mock(server_info='mock_peer')
        loc_header = {'block_height': 2020, 'block_hash': 'ff' * 32}
        net_header = {'block_height': 2021, 'block_hash': 'aa' * 32, 'prev_block_hash': 'ff'*32, 'header_bytes': b''}
        net_header_b = {'block_height': 2021, 'block_hash': 'bb' * 32, 'prev_block_hash': 'ff' * 32, 'header_bytes': b''}
        self.interface.get_header.side_effect = [async_coro(net_header_b), async_coro(net_header)]

        self.repo.get_best_header.return_value = loc_header

        self.loop.run_until_complete(self.sut.on_new_header(peer, net_header))
        self.assertEqual(self.repo.get_best_header.call_count, 1)

        self.assertEqual(0, len(self.interface.method_calls))
        self.assertEqual(0, len(self.electrod_loop.method_calls))
        self.assertEqual(2, len(self.repo.method_calls))

    def test_connectionpool_offline(self):
        self.interface.is_pool_online = False
        self.loop.run_until_complete(self.sut.check_headers())
        Mock.assert_called_with(self.delay_task_runner, coro_call('check_headers'),
                                self.sut.new_headers_fallback_poll_interval)



import asyncio
import gc
from spruned.application.logging_factory import Logger
from spruned.application.tools import async_delayed_task
from spruned.builder import cache, headers_reactor, blocks_reactor, jsonrpc_server, repository


async def main_task(loop):
    try:
        Logger.leveldb.debug('Ensuring integrity of the storage, and tracking missing items')
        try:
            await loop_check_integrity(loop)
        except asyncio.TimeoutError:
            Logger.cache.error('There must be an error in storage, 30 seconds to check are too many')
        Logger.leveldb.debug('Checking cache limits')
        try:
            asyncio.wait_for(asyncio.gather(cache.check()), timeout=30)
        except asyncio.TimeoutError:
            Logger.cache.error('There must be an error in cache, 30 seconds to check are too many')
        headers_reactor.add_on_best_height_hit_callbacks(blocks_reactor.start())
        headers_reactor.add_on_best_height_hit_callbacks(blocks_reactor.bootstrap_blocks())
        loop.create_task(headers_reactor.start())
        loop.create_task(jsonrpc_server.start())
        loop.create_task(cache.lurk())
        loop.create_task(loop_collect_garbage(loop))
    finally:
        pass


async def loop_check_integrity(l):
    """
    this task also prune blocks
    """
    await repository.ensure_integrity()


async def loop_collect_garbage(l):
    gc.collect()
    l.create_task(async_delayed_task(loop_collect_garbage(l), 300))



import abc
import asyncio
import random
import time
from typing import List

from spruned.application.logging_factory import Logger
from spruned.application.tools import check_internet_connection, async_delayed_task
from spruned.daemon import exceptions
from spruned.daemon.abstracts import ConnectionPoolAbstract, ConnectionAbstract


class BaseConnectionPool(ConnectionPoolAbstract, metaclass=abc.ABCMeta):
    def __init__(self,
                 peers=list(),
                 network_checker=check_internet_connection,
                 delayer=async_delayed_task,
                 loop=asyncio.get_event_loop(),
                 use_tor=False,
                 connections=3,
                 sleep_no_internet=30,
                 ipv6=False
                 ):
        self._connections = []
        self._peers = peers
        self._headers_observers = []
        self._new_peers_observers = []
        self._on_connect_observers = []
        self._required_connections = connections
        self._network_checker = network_checker
        self._use_tor = use_tor
        self.loop = loop
        self.delayer = delayer
        self._connection_notified = False
        self._is_online = False
        self._sleep_on_no_internet_connectivity = sleep_no_internet
        self._keepalive = True
        self._ipv6 = ipv6

    @property
    def peers(self):
        return self._peers

    @property
    def connections(self):
        connections = []
        for c in self._connections:
            if c.connected and c.score >= 0:
                connections.append(c)
        return connections

    @property
    def established_connections(self):
        return [connection for connection in self.connections if connection.connected]

    def _pick_peer(self):
        i = 0
        while 1:
            if self.peers:
                server = random.choice(self.peers)
                if server not in [connection.hostname for connection in self.connections]:
                    if ':' in server[0] and not self._ipv6:
                        i += 1
                        continue
                    return server
                i += 1
                if i < 100:
                    continue
            raise exceptions.NoServersException

    def _pick_multiple_peers(self, howmany: int):
        assert howmany >= 1
        i = 0
        servers = []
        while 1:
            if self.peers:
                server = self._pick_peer()
                if server in servers:
                    continue
                servers.append(server)
                if len(servers) == howmany:
                    return servers
                if i < 100:
                    continue
            raise exceptions.NoServersException

    def _pick_connection(self, fail_silent=False):
        i = 0
        while 1:
            if self.established_connections:
                connection = random.choice(self.established_connections)
                if connection.connected and connection.score > 0:
                    return connection
                i += 1
                if i < 100:
                    continue
            if not fail_silent:
                raise exceptions.NoPeersException
            return

    def _pick_multiple_connections(self, howmany: int) -> List[ConnectionAbstract]:
        assert howmany >= 1
        i = 0
        connections = []
        while 1:
            if self.established_connections:
                connection = self._pick_connection()
                if connection in connections:
                    i += 1
                    if i > 100:
                        raise exceptions.NoPeersException
                    continue
                connections.append(connection)
                if len(connections) == howmany:
                    return connections
            i += 1
            if i < 100:
                continue
            raise exceptions.NoPeersException

    def is_online(self):
        return self._is_online

    def add_on_connected_observer(self, observer):
        self._on_connect_observers.append(observer)

    def add_header_observer(self, observer):
        self._headers_observers.append(observer)

    def on_peer_disconnected(self, peer: ConnectionAbstract, *_):
        peer.add_error(int(time.time()) + 180)

    async def on_peer_received_header(self, peer: ConnectionAbstract, *_):
        for observer in self._headers_observers:
            self.loop.create_task(self.delayer(observer(peer, peer.last_header)))

    async def on_peer_received_peers(self, peer: ConnectionAbstract, *_):
        raise NotImplementedError

    async def on_peer_error(self, peer: ConnectionAbstract, error_type=None):
        if error_type == 'connect':
            if await self._check_internet_connectivity():
                peer.add_error(int(time.time()) + 180)
            return
        if self.is_online:
            Logger.electrum.debug('Peer %s error', peer)
            await self._handle_peer_error(peer)

    def stop(self):
        self._keepalive = False

    async def _check_internet_connectivity(self):
        if self._network_checker is None:  # pragma: no cover
            self._is_online = True
            return
        self._is_online = self._network_checker()
        return self._is_online

    async def _handle_peer_error(self, peer: ConnectionAbstract):
        Logger.electrum.debug('Handling connection error for %s', peer.hostname)
        if not peer.connected:
            peer.add_error()
            return
        if not peer.score:
            Logger.electrum.error('Disconnecting from peer %s, score: %s', peer.hostname, peer.score)
            self.loop.create_task(self.delayer(peer.disconnect()))
            return
        if not await peer.ping(timeout=2):
            Logger.electrum.error('Ping timeout from peer %s, score: %s', peer.hostname, peer.score)
            self.loop.create_task(self.delayer(peer.disconnect()))

    def connect(self):
        raise NotImplementedError



import asyncio
import base64
import binascii
import gc
from aiohttp import web
from jsonrpcserver.aio import methods
from jsonrpcserver import config, status
from jsonrpcserver.exceptions import JsonRpcServerError
from jsonrpcserver.response import ExceptionResponse

from spruned.application.logging_factory import Logger
from spruned.application.tools import async_delayed_task
from spruned.daemon.exceptions import GenesisTransactionRequestedException
from spruned import __version__ as spruned_version
from spruned import __bitcoind_version_emulation__ as bitcoind_version

config.schema_validation = False

API_HELP = """
spruned %s, emulating bitcoind %s

== Blockchain ==
getbestblockhash
getblock "blockhash" ( verbosity ) 
getblockchaininfo
getblockcount
getblockhash height
getblockheader "hash" ( verbose )
gettxout "txid" n ( include_mempool )

== Rawtransactions ==
getrawtransaction "txid" ( verbose )
sendrawtransaction "hexstring" ( allowhighfees )

== Util ==
estimatefee nblocks
estimatesmartfee conf_target ("estimate_mode")

""" % (spruned_version, bitcoind_version)


class JsonRpcServerException(JsonRpcServerError):
    def __init__(self, code, message, data=None):
        super().__init__(data=data)
        self.code = code
        self.message = message
        self.http_status = status.HTTP_BAD_REQUEST


class JSONRPCServer:
    def __init__(self, host, port, username, password):
        self.username = username.encode()
        self.password = password.encode()
        self.host = host
        self.port = port
        self.vo_service = None
        self._auth = 'Basic %s' % base64.b64encode(self.username + b':' + self.password).decode()

    def set_vo_service(self, vo_service):
        self.vo_service = vo_service

    def _authenticate(self, request):
        return bool(request.headers.get('Authorization') == self._auth)

    async def _handle(self, jsonrequest):
        if not self._authenticate(jsonrequest):
            return web.json_response({}, status=401)
        request = await jsonrequest.json()
        result = {
            "id": request.get("id"),
            "result": None,
            "error": None
        }
        response = await methods.dispatch(request)
        if isinstance(response, ExceptionResponse):
            response['result'] = response.get('result', None)
            return web.json_response(response, status=response.http_status)
        result.update(response)
        return web.json_response(result)

    async def start(self):
        app = web.Application()
        app.router.add_post('/', self._handle)
        runner = web.AppRunner(app)
        await runner.setup()
        methods.add(self.echo)
        methods.add(self.help)
        methods.add(self.estimatefee)
        methods.add(self.estimatesmartfee)
        methods.add(self.getbestblockhash)
        methods.add(self.getblockchaininfo)
        methods.add(self.getblockheader)
        methods.add(self.getblockhash)
        methods.add(self.getblock)
        methods.add(self.getblockcount)
        methods.add(self.getrawtransaction)
        methods.add(self.gettxout)
        methods.add(self.dev_memorysummary, name="dev-gc-stats")
        methods.add(self.dev_collect, name="dev-gc-collect")
        methods.add(self.stop, name="stop")
        methods.add(self.sendrawtransaction)

        return await web.TCPSite(runner, host=self.host, port=self.port).start()

    async def help(self, *args):
        return API_HELP

    async def echo(self, *args):
        return ""

    async def getblock(self, blockhash: str, mode: int = 1):
        try:
            blockhash = blockhash.strip()
            binascii.unhexlify(blockhash)
            assert len(blockhash) == 64
        except (binascii.Error, AssertionError):
            raise JsonRpcServerException(
                code=-5,
                message="Error parsing JSON:%s" % blockhash
            )
        response = await self.vo_service.getblock(blockhash, mode)
        if not response:
            raise JsonRpcServerException(code=-5, message="Block not found")
        return response

    async def getrawtransaction(self, txid: str, verbose=False):
        if verbose:
            raise JsonRpcServerException(
                code=-8,
                message="spruned error: Verbose mode not supported"
            )
        try:
            txid = txid.strip()
            binascii.unhexlify(txid)
        except binascii.Error:
            raise JsonRpcServerException(
                code=-8,
                message="parameter 1 must be hexadecimal string (not '%s')" % txid
            )
        if len(txid) != 64:
            raise JsonRpcServerException(
                code=-8,
                message="parameter 1 must be of length 64 (not '%s')" % len(txid)
            )
        try:
            response = await self.vo_service.getrawtransaction(txid, verbose)
        except GenesisTransactionRequestedException:
            raise JsonRpcServerException(
                code=-5,
                message="The genesis block coinbase is not considered an ordinary transaction and cannot be retrieved"
            )
        if not response:
            raise JsonRpcServerException(
                code=-5,
                message="No such mempool or blockchain transaction. [maybe try again]"
            )
        return response

    async def getbestblockhash(self):
        return await self.vo_service.getbestblockhash()

    async def sendrawtransaction(self, rawtx: str):
        try:
            binascii.unhexlify(rawtx)
        except (binascii.Error, AssertionError):
            raise JsonRpcServerException(
                code=-22,
                message="TX decode failed"
            )
        return await self.vo_service.sendrawtransaction(rawtx)

    async def getblockcount(self):
        res = await self.vo_service.getblockcount()
        return res

    async def getblockhash(self, blockheight: int):
        try:
            int(blockheight)
        except ValueError:
            raise JsonRpcServerException(
                code=-5,
                message="Error parsing JSON:%s" % blockheight
            )
        response = await self.vo_service.getblockhash(blockheight)
        if not response:
            raise JsonRpcServerException(
                code=-5,
                message="Block height out of range"
            )
        return response

    async def getblockheader(self, blockhash: str, verbose=True):
        try:
            blockhash = blockhash.strip()
            binascii.unhexlify(blockhash)
            assert len(blockhash) == 64
        except (binascii.Error, AssertionError):
            raise JsonRpcServerException(
                code=-5,
                message="Error parsing JSON:%s" % blockhash
            )
        response = await self.vo_service.getblockheader(blockhash, verbose=verbose)
        if not response:
            raise JsonRpcServerException(
                code=-5,
                message="Block not found"
            )
        return response

    async def estimatefee(self, blocks: int):
        try:
            int(blocks)
        except ValueError:
            raise JsonRpcServerException(
                code=-5,
                message="Error parsing JSON:%s" % blocks
            )
        response = await self.vo_service.estimatefee(blocks)
        if response is None:
            return "-1"
        return response

    async def estimatesmartfee(self, blocks: int, estimate_mode=None):
        try:
            int(blocks)
        except ValueError:
            raise JsonRpcServerException(
                code=-5,
                message="Error parsing JSON:%s" % blocks
            )
        if not 0 < int(blocks) < 1009:
            raise JsonRpcServerException(
                code=-8,
                message="Invalid conf_target, must be between 1 - 1008"
            )
        response = await self.vo_service.estimatefee(blocks)
        if response is None:
            raise JsonRpcServerException(
                code=-8,
                message="server error: try again"
            )
        return {
            "blocks": blocks,
            "feerate": response,
            "_feerate": "{:.8f}".format(response)
        }

    async def getblockchaininfo(self):
        response = await self.vo_service.getblockchaininfo()
        if response is None:
            raise JsonRpcServerException(
                code=-8,
                message="server error: try again"
            )
        return response

    async def gettxout(self, txid: str, index: int):
        try:
            txid = txid.strip()
            response = await self.vo_service.gettxout(txid, index)
        except:
            Logger.jsonrpc.error('Error in gettxout', exc_info=True)
            raise JsonRpcServerException(
                code=-8,
                message="server error: try again"
            )
        return response

    async def dev_memorysummary(self):
        return {"stats": gc.get_stats()}

    async def dev_collect(self):
        res = {
            "before": gc.get_stats()
        }
        gc.collect()
        res['after'] = gc.get_stats()
        return res

    async def stop(self):
        loop = asyncio.get_event_loop()

        async def stop():
            loop.stop()
            loop.close()
        loop.create_task(async_delayed_task(stop(), 0))
        return {"error": None, "response": None}



import asyncio
from unittest import TestCase
from unittest.mock import Mock

from spruned.repositories.blockchain_repository import BLOCK_PREFIX
from spruned.repositories.repository import Repository


class TestRepository(TestCase):
    def setUp(self):
        self.headers = Mock()
        self.blocks = Mock()
        self.cache = Mock(cache_name=b'cache_prefix')
        self.sut = Repository(self.headers, self.blocks, keep_blocks=5)
        self.sut.set_cache(self.cache)
        self.sut.ldb = Mock()
        self.loop = asyncio.get_event_loop()

    def tearDown(self):
        self.headers.reset_mock()
        self.blocks.reset_mock()
        self.cache.reset_mock()
        self.sut.ldb.reset_mock()

    def test_check_stales(self):
        self.headers.get_best_header.return_value = {'block_height': 16, 'block_hash': 'block10'}
        self.headers.get_headers_since_height.return_value = [
            {'block_height': 12, 'block_hash': 'block12'},
            {'block_height': 13, 'block_hash': 'block13'},
            {'block_height': 14, 'block_hash': 'block14'},
            {'block_height': 15, 'block_hash': 'block15'},
            {'block_height': 16, 'block_hash': 'block16'}
        ]
        self.blocks.storage_name = b'block_prefix'
        self.blocks.get_key.side_effect = lambda x, y: b'block_prefix.' + BLOCK_PREFIX + b'.' + x.encode()
        self.cache.get_index.return_value = {
            'keys': {
                b'block12': {},
                b'block13': {},
                b'block14': {},
                b'block15': {}
            }
        }
        iterator = [
            (b'block_prefix.block12',),
            (b'block_prefix.block13',),
            (b'block_prefix.block14',),
            (b'block_prefix.block15',),
            (b'block_prefix.block16',),
        ]
        self.sut.ldb.iterator.return_value = iterator
        self.loop.run_until_complete(self.sut.ensure_integrity())
        Mock.assert_called_once_with(self.blocks.remove_block, b'block_prefix.block16')

    def test_check_stales_no_index(self):
        self.headers.get_best_header.return_value = {'block_height': 16, 'block_hash': 'block10'}
        self.cache.get_index.return_value = None
        self.headers.get_headers_since_height.return_value = [
            {'block_height': 12, 'block_hash': 'block12'},
            {'block_height': 13, 'block_hash': 'block13'},
            {'block_height': 14, 'block_hash': 'block14'},
            {'block_height': 15, 'block_hash': 'block15'},
            {'block_height': 16, 'block_hash': 'block16'}
        ]
        self.blocks.storage_name = b'block_prefix'
        self.blocks.get_key.side_effect = lambda x, y: b'block_prefix.' + BLOCK_PREFIX + b'.' + x.encode()
        self.cache.get_index.return_value = None
        Mock.assert_not_called(self.blocks.remove_block)



import asyncio
import random
from unittest import TestCase
from unittest.mock import Mock, call
from spruned.application.jsonrpc_server import JSONRPCServer
from spruned.application.utils.jsonrpc_client import JSONClient
from test.utils import async_coro


class TestJSONRPCServerGetblock(TestCase):
    def setUp(self):
        bindport = random.randint(31337, 41337)
        self.sut = JSONRPCServer('127.0.0.1', bindport, 'testuser', 'testpassword')
        self.vo_service = Mock()
        self.sut.set_vo_service(self.vo_service)
        self.client = JSONClient(b'testuser', b'testpassword', '127.0.0.1', bindport)
        self.loop = asyncio.get_event_loop()

    def test_getblock_success(self):
        self.vo_service.getblock.side_effect = [async_coro({'block': 'header'}),
                                                      async_coro('cafebabe')]

        async def test():
            await self.sut.start()
            response = await self.client.call('getblock', params=['00' * 32])
            response2 = await self.client.call('getblock', params=['00' * 32, False])
            return response, response2

        res, res2 = self.loop.run_until_complete(test())
        self.assertEqual(
            res,
            {'error': None, 'id': 1, 'jsonrpc': '2.0', 'result': {'block': 'header'}}
        )
        self.assertEqual(
            res2,
            {'error': None, 'id': 1, 'jsonrpc': '2.0', 'result': 'cafebabe'}
        )
        Mock.assert_has_calls(
            self.vo_service.getblock,
            calls=[
                call('00' * 32, 1),
                call('00' * 32, False)
            ]
        )

    def test_getblock_error_missing(self):
        response = None
        self.vo_service.getblock.return_value = async_coro(response)

        async def test():
            await self.sut.start()
            response = await self.client.call('getblock', params=['00'*32])
            return response

        res = self.loop.run_until_complete(test())
        self.assertEqual(
            res,
            {'error': {'code': -5, 'message': 'Block not found'}, 'id': 1, 'jsonrpc': '2.0', 'result': None}
        )
        Mock.assert_called_with(self.vo_service.getblock, '00' * 32, 1)

    def test_getblock_error_error_params(self):
        response = None
        self.vo_service.getblock.return_value = async_coro(response)

        async def test():
            await self.sut.start()
            response1 = await self.client.call('getblock', params=['wrong_blockhash'])
            response2 = await self.client.call('getblock')
            return response1, response2

        res, res2 = self.loop.run_until_complete(test())
        self.assertEqual(
            res,
            {
                'jsonrpc': '2.0',
                'error': {
                    'code': -5, 'message': 'Error parsing JSON:wrong_blockhash'
                },
                'id': 1,
                'result': None
            }
        )
        self.assertEqual(
            res2,
            {'jsonrpc': '2.0', 'error': {'code': -32602, 'message': 'Invalid params'}, 'id': 1, 'result': None}
        )
        Mock.assert_not_called(self.vo_service.getblock)



import unittest

from spruned import settings
from spruned.repositories.headers_repository import HeadersSQLiteRepository
from spruned.application.database import sqlite
from spruned.daemon import exceptions
from test.utils import make_headers


class TestHeadersRepository(unittest.TestCase):
    def setUp(self):
        assert not settings.SQLITE_DBNAME
        self.sut = HeadersSQLiteRepository(sqlite)

    def tests_headers_repository_ok(self):
        """
        all the successful headers repository calls can be made.
        """
        self.sut.remove_headers_after_height(0)
        headers = make_headers(0, 5, with_timestamp=False)
        res = self.sut.save_header(
            headers[0]['block_hash'],
            headers[0]['block_height'],
            headers[0]['header_bytes'],
            None
        )
        self.assertEqual(res, headers[0])
        self.sut.save_headers(headers[1:])
        best_header = self.sut.get_best_header()
        self.assertEqual(best_header, headers[-1])
        self.assertEqual(self.sut.get_header_at_height(4), headers[4])
        self.assertEqual(self.sut.get_block_hash(3), headers[3]['block_hash'])
        self.assertEqual(self.sut.get_block_height(headers[3]['block_hash']), 3)
        header3 = self.sut.get_header_at_height(3)
        self.assertEqual(header3.pop('next_block_hash'), headers[4]['block_hash'])
        self.assertEqual(header3, headers[3])

        headers2 = make_headers(5, 10, with_timestamp=False, prevblock_0=headers[4]['block_hash'])

        res = self.sut.save_header(
            headers2[0]['block_hash'],
            headers2[0]['block_height'],
            headers2[0]['header_bytes'],
            headers2[0]['prev_block_hash']
        )
        self.assertEqual(res, headers2[0])
        self.sut.save_headers(headers2[1:])
        self.assertEqual(self.sut.get_best_header(), headers2[-1])

        self.sut.remove_headers_after_height(headers2[0]['block_height'])
        self.assertEqual(self.sut.get_best_header(), headers[-1])

        self.sut.save_headers(headers2)
        self.assertEqual(self.sut.get_best_header(), headers2[-1])

        self.sut.remove_header_at_height(headers2[-1]['block_height'])
        res = self.sut.save_header(
            headers2[-1]['block_hash'],
            headers2[-1]['block_height'],
            headers2[-1]['header_bytes'],
            headers2[-1]['prev_block_hash']
        )
        self.assertEqual(res, headers2[-1])
        self.assertEqual(self.sut.get_best_header(), headers2[-1])

        res = self.sut.get_headers_since_height(headers2[3]['block_height'])
        self.assertEqual(res[0].pop('next_block_hash'), headers2[4]['block_hash'])
        self.assertEqual(headers2[3:], res)

    def test_headers_repository_save_wrong_headers(self):
        self.sut.remove_headers_after_height(0)
        headers = make_headers(0, 5, with_timestamp=False)
        res = self.sut.save_header(
            headers[0]['block_hash'],
            headers[0]['block_height'],
            headers[0]['header_bytes'],
            None
        )
        self.assertEqual(res, headers[0])
        self.sut.save_headers(headers[1:])
        headers2 = make_headers(5, 6, with_timestamp=False, prevblock_0=headers[3]['block_hash'])
        with self.assertRaises(exceptions.HeadersInconsistencyException):
            self.sut.save_header(
                headers2[0]['block_hash'],
                headers2[0]['block_height'],
                headers2[0]['header_bytes'],
                headers2[0]['prev_block_hash']
            )



from typing import Dict

from spruned.daemon.bitcoin_p2p import utils
from spruned.daemon.bitcoin_p2p.p2p_connection import P2PConnectionPool
from spruned.daemon.bitcoin_p2p.p2p_interface import P2PInterface


def build(network: Dict):
    assert isinstance(network, dict), network
    pool = P2PConnectionPool(connections=8, batcher_timeout=30, network=network['pycoin'])
    interface = P2PInterface(pool, network=network['pycoin'])
    return pool, interface



import asyncio
from typing import Dict, List
from pycoin.message.InvItem import InvItem, ITEM_TYPE_SEGWIT_BLOCK
from pycoin.serialize import h2b_rev
from pycoinnet.networks import MAINNET

from spruned.application import exceptions
from spruned.daemon.bitcoin_p2p import utils
from spruned.daemon.bitcoin_p2p.p2p_connection import P2PConnectionPool


class P2PInterface:
    def __init__(self,
                 connection_pool: P2PConnectionPool, loop=asyncio.get_event_loop(),
                 network=MAINNET, peers_bootstrapper=utils.dns_bootstrap_servers):
        self.pool = connection_pool
        self._on_connect_callbacks = []
        self.loop = loop
        self.network = network
        self._bootstrap_status = 0
        self.peers_bootstrapper = peers_bootstrapper

    async def on_connect(self):
        for callback in self._on_connect_callbacks:
            self.loop.create_task(callback())

    async def get_block(self, blockhash: str, peers=None, timeout=None) -> Dict:
        inv_item = InvItem(ITEM_TYPE_SEGWIT_BLOCK, h2b_rev(blockhash))
        response = await self.pool.get(inv_item, peers=peers, timeout=timeout)
        return response and {
            "block_hash": str(response.hash()),
            "prev_block_hash": str(response.previous_block_hash),
            "timestamp": int(response.timestamp),
            "header_bytes": bytes(response.as_blockheader().as_bin()),
            "block_object": response,
            "block_bytes": bytes(response.as_bin())
        }

    async def get_blocks(self, *blockhash: str) -> Dict:
        """
        I have to work on pycoinnet to understand how the invbatcher can handle a more efficient 'getblocks'.
        Meanwhile, parallelize getblocks. This may be dirty, let's try it....
        """
        sorted_hash = [x for x in blockhash]
        blocks = {}
        r, max_retry = 0, 100
        while sorted_hash:
            r += 1
            if r > max_retry:
                raise ValueError
            _blocks = await asyncio.gather(*(self.get_block(h) for h in sorted_hash), return_exceptions=True)
            for i, _hash in enumerate(sorted_hash):
                if isinstance(_blocks[i], dict):
                    blocks[_hash] = _blocks[i]
            sorted_hash = [h for h in sorted_hash if h not in blocks]
        return blocks

    def add_on_connect_callback(self, callback):
        self._on_connect_callbacks.append(callback)

    async def start(self):
        self.pool.add_on_connected_observer(self.on_connect)
        peers = None
        i = 0
        while not peers:
            if i > 10:
                raise exceptions.SprunedException
            peers = await self.peers_bootstrapper(self.network)
            i += 1
        _ = [self.pool.add_peer(peer) for peer in peers]
        self.loop.create_task(self.pool.connect())

    def set_bootstrap_status(self, value: float):
        self._bootstrap_status = value

    @property
    def bootstrap_status(self):
        return self._bootstrap_status



#!/usr/bin/env python3
# Copyright (C) 2018 Guido Dassori <guido.dassori@gmail.com>
#

import sys
if sys.version < '3.5.2':
    raise ValueError('Python >= 3.5.2 is required')
sys.path.insert(0, './')

import argparse
import asyncio
from daemonize import Daemonize
from spruned.application.context import ctx
from spruned.application.tools import create_directory




parser = argparse.ArgumentParser(
    description="A Bitcoin Lightweight Pseudonode",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter
)
parser.add_argument(
    '--rpcuser',
    action='store', dest='rpcuser', default=ctx.rpcuser,
    help='Username for JSON-RPC connections'
)
parser.add_argument(
    '--rpcpassword',
    action='store', dest='rpcpassword', default=ctx.rpcpassword,
    help='Password for JSON-RPC connections'
)
parser.add_argument(
    '--rpcport',
    action='store', dest='rpcport', default=ctx.rpcport,
    help='Listen for JSON-RPC connections on <port> (default: 8332 or testnet: 18332)'
)
parser.add_argument(
    '--rpcbind',
    action='store', dest='rpcbind', default=ctx.rpcbind,
    help='Bind to given address to listen for JSON-RPC connections.'
)
parser.add_argument(
    '--datadir',
    action='store', dest='datadir', default=ctx.datadir,
    help='Specify data directory'
)
parser.add_argument(
    '--daemon',
    action='store_true', dest='daemonize', default=bool(ctx.daemonize),
    help='Run in the background as a daemon and accept commands'
)
parser.add_argument(
    '--keep-blocks',
    action='store', dest='keep_blocks', default=int(ctx.keep_blocks),
    help='', type=int
)
parser.add_argument(
    '--network',
    action='store', dest='network',
    choices=[
        'bitcoin.mainnet',
        'bitcoin.testnet',
        #'bitcoin.regtest'
    ],
    help=''
)
parser.add_argument(
    '--debug',
    action='store_true', dest='debug', default=ctx.debug,
    help='Enable debug mode'
)
parser.add_argument(
    '--cache-size',
    action='store', dest='cache_size', default=int(ctx.cache_size),
    help='Cache size (in megabytes)'
)


def main():   # pragma: no cover

    def start():  # pragma: no cover
        ctx.load_args(args)
        create_directory()
        main_loop = asyncio.get_event_loop()
        from spruned.main import main_task
        main_loop.create_task(main_task(main_loop))
        main_loop.run_forever()

    args = parser.parse_args()
    if args.daemonize:
        pid = ctx.datadir + 'spruned.pid'
        daemon = Daemonize(app="spruned", pid=pid, action=start)
        daemon.start()
    else:
        start()


if __name__ == '__main__':  # pragma: no cover
    main()



import asyncio
import io
import binascii
from bitcoin import deserialize
from pycoin.block import Block
from spruned import settings
from spruned.application.cache import CacheAgent
from spruned.application.logging_factory import Logger
from spruned.application.tools import deserialize_header, script_to_scripthash
from spruned.application import exceptions
from spruned.application.abstracts import RPCAPIService
from spruned.daemon.exceptions import ElectrodMissingResponseException, GenesisTransactionRequestedException


class SprunedVOService(RPCAPIService):
    def __init__(self, electrod, p2p, cache: CacheAgent=None, utxo_tracker=None, repository=None,
                 loop=asyncio.get_event_loop()):
        self.sources = []
        self.primary = []
        self.cache = cache
        self.p2p = p2p
        self.electrod = electrod
        self.min_sources = 1
        self.current_best_height = None
        self.utxo_tracker = utxo_tracker
        self.repository = repository
        self.loop = loop
        self._last_estimatefee = None

    def available(self):
        raise NotImplementedError

    async def getblock(self, blockhash: str, mode: int=1):
        block_header = self.repository.headers.get_block_header(blockhash)
        if not block_header:
            return
        block = await self._get_block(block_header)
        if mode == 1:
            block_object = block.get(
                'block_object',
                Block.parse(io.BytesIO(block['block_bytes']), check_merkle_hash=False)
            )
            best_header = self.repository.headers.get_best_header()
            block['confirmations'] = best_header['block_height'] - block_header['block_height']
            serialized = self._serialize_header(block_header)
            serialized['tx'] = [tx.id() for tx in block_object.txs]
            del block
            return serialized
        elif mode == 2:
            raise NotImplementedError
        bb = block['block_bytes']
        del block
        return binascii.hexlify(bb).decode()

    async def _get_block(self, blockheader, _r=0):
        blockhash = blockheader['block_hash']
        storedblock = self.repository.blockchain.get_block(blockhash)
        block = storedblock or await self.p2p.get_block(blockhash, timeout=10)
        if not block:
            if _r > 10:
                raise exceptions.ServiceException
            else:
                return await self._get_block(blockheader, _r + 1)
        if not storedblock:
            self.loop.create_task(self.repository.blockchain.async_save_block(block, tracker=self.cache))
        return block

    async def getrawtransaction(self, txid: str, verbose=False):
        repo_tx = self.repository.blockchain.get_transaction(txid)
        transaction = repo_tx or await self.electrod.getrawtransaction(txid)
        res = {
            'source': 'p2p' if repo_tx else 'electrum',
            'rawtx': binascii.hexlify(repo_tx['transaction_object'].as_bin()).decode() if repo_tx else transaction
        }
        if verbose:
            return res
        return res['rawtx']

    async def getbestblockhash(self):
        res = self.repository.headers.get_best_header().get('block_hash')
        return res and res

    async def sendrawtransaction(self, rawtx: str):
        res = await self.electrod.sendrawtransaction(rawtx)
        return res

    async def getblockhash(self, blockheight: int):
        return self.repository.headers.get_block_hash(blockheight)

    async def getblockheader(self, blockhash: str, verbose=True):
        header = self.repository.headers.get_block_header(blockhash)
        if verbose:
            _best_header = self.repository.headers.get_best_header()
            res = self._serialize_header(header)
            res["confirmations"] = _best_header['block_height'] - header['block_height'] + 1
        else:
            res = binascii.hexlify(header['header_bytes']).decode()
        return res

    @staticmethod
    def _serialize_header(header):
        _deserialized_header = deserialize_header(binascii.hexlify(header['header_bytes']).decode())
        return {
            "hash": _deserialized_header['hash'],
            "height": header['block_height'],
            "version": _deserialized_header['version'],
            "versionHex": "Not Implemented Yet",
            "merkleroot": _deserialized_header['merkle_root'],
            "time": _deserialized_header['timestamp'],
            "mediantime": _deserialized_header['timestamp'],
            "nonce": _deserialized_header['nonce'],
            "bits": _deserialized_header['bits'],
            "difficulty": "Not Implemented Yet",
            "chainwork": "Not Implemented Yet",
            "previousblockhash": _deserialized_header['prev_block_hash'],
            "nextblockhash": header.get('next_block_hash')
        }

    async def getblockcount(self):
        return self.repository.headers.get_best_header().get('block_height')

    async def estimatefee(self, blocks: int):
        try:
            self._last_estimatefee = await self._estimatefee(blocks)
        except:
            pass
        return self._last_estimatefee

    async def _estimatefee(self, blocks, _r=1):
        try:
            res = await self.electrod.estimatefee(blocks)
        except ElectrodMissingResponseException as e:
            Logger.electrum.error('Error with peer', exc_info=True)
            _r += 1
            if _r > 5:
                raise e
            return await self._estimatefee(blocks, _r + 1)
        return res

    async def getbestblockheader(self, verbose=True):
        best_header = self.repository.headers.get_best_header()
        return await self.getblockheader(best_header['block_hash'], verbose=verbose)

    async def getblockchaininfo(self):
        from spruned import __version__ as spruned_version
        from spruned import __bitcoind_version_emulation__ as bitcoind_version
        best_header = self.repository.headers.get_best_header()
        _deserialized_header = deserialize_header(best_header['header_bytes'])
        return {
            "chain": "main",
            "warning": "spruned %s, emulating bitcoind v%s" % (spruned_version, bitcoind_version),
            "blocks": best_header["block_height"],
            "headers": best_header["block_height"],
            "bestblockhash": best_header["block_hash"],
            "difficulty": None,
            "chainwork": None,
            "mediantime": _deserialized_header["timestamp"],
            "verificationprogress": self.p2p.bootstrap_status,
            "pruned": False,
        }

    async def gettxout(self, txid: str, index: int):
        repo_tx = self.repository.blockchain.get_transaction(txid)
        transaction = repo_tx and binascii.hexlify(repo_tx['transaction_bytes']).decode() \
                        or await self.electrod.getrawtransaction(txid)
        if not transaction:
            return
        deserialized = deserialize(transaction)
        if index + 1 > len(deserialized['outs']):
            return
        vout = deserialized['outs'][index]
        scripthash = script_to_scripthash(vout['script'])
        unspents = await self._listunspent_by_scripthash(scripthash) or []
        txout = None
        for unspent in unspents:
            if unspent['tx_hash'] == txid and unspent['tx_pos'] == index:
                txout = unspent
        return txout and await self._format_gettxout(txout, vout)

    async def _format_gettxout(self, txout: dict, deserialized_vout: dict):
        best_header = self.repository.headers.get_best_header()
        return {
            "bestblock": best_header['block_hash'],
            "confirmations": best_header['block_height'] - txout['height'],
            "value": "{:.8f}".format(txout['value'] / 10**8, 8),
            "scriptPubKey": {
                "asm": "",  # todo
                "hex": deserialized_vout['script'],
                "reqSigs": 0,  # todo
                "type": "",
                "addresses": []  # todo
            }
        }

    async def _listunspent_by_scripthash(self, scripthash, _r=0):
        try:
            unspents = await self.electrod.listunspents_by_scripthash(scripthash)
        except:
            if _r > 15:
                return
            return await self._listunspent_by_scripthash(scripthash, _r=_r+1)
        return unspents



import asyncio
import base64
import time
import typing
from json import JSONDecodeError

import aiohttp
import json


class JSONClient:
    def __init__(self, user, password, host, port):
        self.url = "http://{}:{}".format(host, port)
        self._auth = 'Basic %s' % base64.b64encode(user + b':' + password).decode()

    async def call(self, method: str, params: typing.List=None, jsonRes=True):
        payload = {
            "method": method,
            "params": params,
            "jsonrpc": "2.0",
            "id": 1,
        }
        async with aiohttp.ClientSession(conn_timeout=10) as session:
            start = time.time()
            response = await session.post(
                self.url,
                data=json.dumps(payload),
                headers={'content-type': 'application/json', 'Authorization': self._auth},
            )
        if jsonRes:
            try:
                return (await response.json())
            except JSONDecodeError as e:
                raise e
        else:
            return response.content


async def getblock_test(cli, bestheight=50000):
    blhash = None
    while not blhash:
        blhash = await cli._call('getblockhash', [bestheight])
        if not blhash:
            await asyncio.sleep(1)
        else:
            raise ValueError
    while 1:
        block = await cli._call('getblock', [blhash])
        if block:
            blhash = block['previousblockhash']
        else:
            await asyncio.sleep(5)

if __name__ == '__main__':
    cli = JSONClient(b'rpcuser', b'passw0rd', '10.8.10.217', 8332)
    loop = asyncio.get_event_loop()
    bestheight = 500000
    print(loop.run_until_complete(cli._call('getblockcount')))
    cli = JSONClient(b'rpcuser', b'rpcpassword', 'localhost', 8332)
    loop = asyncio.get_event_loop()
    bestheight = 500000
    print(loop.run_until_complete(cli._call('getblockcount')))
    #loop.run_until_complete(getblock_test(cli, bestheight=bestheight))



import logging
import sys
from spruned import settings
from spruned.application.context import ctx


class LoggingFactory:
    def __init__(self, loglevel=logging.DEBUG, logfile=None, stdout=False):
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        root = logging.getLogger()
        root.setLevel(level=loglevel)

        file_logger = logfile and logging.FileHandler(logfile)
        stdout_logger = stdout and logging.StreamHandler(sys.stdout)

        stdout_logger and stdout_logger.setLevel(loglevel)
        stdout_logger and stdout_logger.setFormatter(formatter)
        stdout_logger and root.addHandler(stdout_logger)

        file_logger and file_logger.setLevel(loglevel)
        file_logger and file_logger.setFormatter(formatter)
        file_logger and root.addHandler(file_logger)

    @property
    def root(self):
        return logging.getLogger('root')

    @property
    def repository(self):
        return logging.getLogger('repository')

    @property
    def third_party(self):
        return logging.getLogger('third_party')

    @property
    def electrum(self):
        return logging.getLogger('electrum')

    @property
    def p2p(self):
        return logging.getLogger('p2p')

    @property
    def leveldb(self):
        return logging.getLogger('leveldb')

    @property
    def bitcoind(self):
        return logging.getLogger('bitcoind')

    @property
    def cache(self):
        return logging.getLogger('cache')

    @property
    def jsonrpc(self):
        return logging.getLogger('jsonrpc')


if settings.TESTING:
    Logger = LoggingFactory(
        logfile=None,
        loglevel=logging.DEBUG,
        stdout=True
    )  # type: LoggingFactory

elif ctx.debug:
    logging.getLogger('jsonrpcserver.dispatcher.response').setLevel(logging.WARNING)
    logging.getLogger('pycoin').setLevel(logging.DEBUG)
    logging.getLogger('p2p').setLevel(logging.DEBUG)
    logging.getLogger('connectrum').setLevel(logging.DEBUG)
    logging.getLogger('electrum').setLevel(logging.DEBUG)
    logging.getLogger('cache').setLevel(logging.DEBUG)
    logging.getLogger('leveldb').setLevel(logging.DEBUG)
    logging.getLogger('asyncio').setLevel(logging.INFO)
    Logger = LoggingFactory(
        logfile=settings.LOGFILE,
        loglevel=logging.DEBUG,
        stdout=True
    )  # type: LoggingFactory

else:
    logging.getLogger('jsonrpcserver.dispatcher.response').setLevel(logging.WARNING)
    logging.getLogger('pycoin').setLevel(logging.ERROR)
    logging.getLogger('p2p').setLevel(logging.INFO)
    logging.getLogger('connectrum').setLevel(logging.ERROR)
    logging.getLogger('electrum').setLevel(logging.INFO)
    logging.getLogger('cache').setLevel(logging.INFO)
    logging.getLogger('leveldb').setLevel(logging.INFO)
    logging.getLogger('asyncio').setLevel(logging.CRITICAL)
    Logger = LoggingFactory(
        logfile=settings.LOGFILE,
        loglevel=logging.INFO,
        stdout=False
    )  # type: LoggingFactory



import asyncio
import unittest
from unittest.mock import Mock, create_autospec, call
import binascii
from spruned.daemon.electrod.electrod_connection import ElectrodConnectionPool, ElectrodConnection
from spruned.daemon.electrod.electrod_interface import ElectrodInterface
from spruned.daemon.exceptions import ElectrodMissingResponseException
from test.utils import async_coro


class TestElectrodInterface(unittest.TestCase):
    def setUp(self):
        self.connectionpool = create_autospec(ElectrodConnectionPool)
        self.electrod_loop = Mock()
        self.sut = ElectrodInterface(self.connectionpool, self.electrod_loop)
        self.loop = asyncio.get_event_loop()
        self.electrum_header = {
            'block_height': 513526,
            'version': 536870912,
            'prev_block_hash': '000000000000000000323a7adf94b5c7df4221778b502176b998c7a70f0152fe',
            'merkle_root': 'a7f92207352feca838df4293b644fe79b3676487f05972ddbc99274ec92ddb97',
            'timestamp': 1521051443,
            'bits': 391481763,
            'nonce': 2209886775
        }
        self.parsed_header = {
                'block_hash': '0000000000000000001e9a7d5bdb53bf19487d50d9e68e04b7254662d693a6ea',
                'block_height': 513526,
                'header_bytes': b'\x00\x00\x00 \xfeR\x01\x0f\xa7\xc7\x98\xb9v!P\x8bw!B\xdf\xc7\xb5\x94\xdfz:2\x00\x00'
                                b'\x00\x00\x00\x00\x00\x00\x00\x97\xdb-\xc9N\'\x99\xbc\xddrY\xf0\x87dg\xb3y\xfeD\xb6'
                                b'\x93B\xdf8\xa8\xec/5\x07"\xf9\xa73g\xa9Z\xa3\x89U\x1772\xb8\x83',
                'prev_block_hash': '000000000000000000323a7adf94b5c7df4221778b502176b998c7a70f0152fe',
                'timestamp': 1521051443
             }

    def tearDown(self):
        self.connectionpool.reset_mock()
        self.electrod_loop.reset_mock()

    def test_get_header_ok(self):
        peer = Mock()
        self.connectionpool.call.return_value = async_coro((peer, self.electrum_header))
        res = self.loop.run_until_complete(self.sut.get_header(513526))
        self.assertEqual(res, self.parsed_header)
        bitcoind_header = "00000020fe52010fa7c798b97621508b772142dfc7b594df7a3a3200000000000000000097db2dc9" \
                          "4e2799bcdd7259f0876467b379fe44b69342df38a8ec2f350722f9a73367a95aa38955173732b883"
        self.assertEqual(binascii.unhexlify(bitcoind_header), self.parsed_header['header_bytes'])

    def test_get_header_invalid_pow(self):
        peer = Mock()
        peer.disconnect.return_value = 'coroutine'
        electrum_header = {k: v for k, v in self.electrum_header.items()}
        electrum_header.update({'bits': 991481763})  # <- target to the moon
        self.connectionpool.call.return_value = async_coro((peer, electrum_header))
        res = self.loop.run_until_complete(self.sut.get_header(513526))
        self.assertIsNone(res)
        Mock.assert_called_once_with(self.electrod_loop.create_task, 'coroutine')

    def test_get_header_checkpoint_failure(self):
        peer = Mock()
        peer.disconnect.return_value = 'coroutine'
        electrum_header = {k: v for k, v in self.electrum_header.items()}
        electrum_header.update({'block_height': 0})  # <- target to the moon
        self.connectionpool.call.return_value = async_coro((peer, electrum_header))
        res = self.loop.run_until_complete(self.sut.get_header(0))
        self.assertIsNone(res)
        Mock.assert_called_once_with(self.electrod_loop.create_task, 'coroutine')
        Mock.assert_called_with(self.connectionpool.call, 'blockchain.block.get_header', 0, get_peer=True)

    def test_subscription_observers(self):
        data = []

        async def callback(peer, header):
            data.append((peer, header))

        peer = Mock()
        observers = []
        self.connectionpool.add_header_observer.side_effect = lambda x: observers.append(x)
        self.sut.add_header_subscribe_callback(callback)
        self.loop.run_until_complete(observers[0](peer, self.electrum_header))
        self.assertEqual(data[0][0], peer)
        self.assertEqual(data[0][1], self.parsed_header)

    def test_subscription_observers_pow_error(self):
        callback = Mock()
        peer = Mock()
        peer.disconnect.return_value = 'coroutine'
        observers = []
        electrum_header = {k: v for k, v in self.electrum_header.items()}
        electrum_header.update({'bits': 991481763})  # <- target to the moon
        self.connectionpool.add_header_observer.side_effect = lambda x: observers.append(x)
        self.connectionpool.call.return_value = async_coro((peer, electrum_header))
        self.sut.add_header_subscribe_callback(callback)
        self.loop.run_until_complete(observers[0](peer, electrum_header))
        res = self.loop.run_until_complete(self.sut.get_header(513526))
        self.assertIsNone(res)
        Mock.assert_called_with(self.electrod_loop.create_task, 'coroutine')

    def test_get_headers_no_response(self):
        self.connectionpool.call.side_effect = [
            ElectrodMissingResponseException, ElectrodMissingResponseException, async_coro(('a', self.electrum_header))
        ]
        with self.assertRaises(ElectrodMissingResponseException):
            self.loop.run_until_complete(self.sut.get_header(10))
        self.assertIsNone(self.loop.run_until_complete(self.sut.get_header(10, fail_silent_out_of_range=True)))
        peer, res = self.loop.run_until_complete(self.sut.get_header(10, get_peer=1))
        self.assertEqual(peer, 'a')
        self.assertEqual(res, self.parsed_header)
        Mock.assert_called_with(self.connectionpool.call, 'blockchain.block.get_header', 10, get_peer=True)

    def test_var_methods(self):
        self.electrod_loop.create_task.return_value = 1
        self.loop.run_until_complete(self.sut.start())
        peer = create_autospec(ElectrodConnection)
        self.loop.run_until_complete(self.sut.disconnect_from_peer(peer))
        Mock.assert_called_with(peer.disconnect)
        self.assertEqual(2, len(self.electrod_loop.method_calls))
        self.electrod_loop.reset_mock()

        self.sut.add_on_connected_callback('callback')
        self.connectionpool.add_on_connected_observer.return_value = True
        Mock.assert_called_with(
            self.connectionpool.add_on_connected_observer,
            'callback'
        )

        self.connectionpool.call.side_effect = [
            async_coro('cafe'),
            async_coro({'hex': 'rawtx'}),
            async_coro('chunk_of_headers'),
            async_coro({'list': 'unspents'}),
            async_coro({'txs': 'list'}),
            async_coro({'fee': 'rate'})
        ]
        peer = Mock()
        self.connectionpool.on_peer_error.return_value = async_coro(lambda x: x.close())
        self.loop.run_until_complete(self.sut.getrawtransaction('cafebabe'))
        self.loop.run_until_complete(self.sut.getrawtransaction('cafebabe', verbose=True))
        self.loop.run_until_complete(self.sut.get_chunk(1))
        self.loop.run_until_complete(self.sut.listunspents_by_address('address'))
        self.loop.run_until_complete(self.sut.getaddresshistory('scripthash'))
        self.loop.run_until_complete(self.sut.estimatefee(6))
        self.loop.run_until_complete(self.sut.handle_peer_error(peer))
        Mock.assert_has_calls(
            self.connectionpool.call,
            calls=[
                call('blockchain.transaction.get', 'cafebabe'),
                call('blockchain.transaction.get', 'cafebabe'),
                call('blockchain.block.get_chunk', 1, get_peer=False),
                call('blockchain.address.listunspent', 'address'),
                call('blockchain.address.get_history', 'scripthash'),
                call('blockchain.estimatefee', 6)
            ]
        )
        Mock.assert_called_with(self.connectionpool.on_peer_error, peer)

    def test_get_headers_in_range(self):

        async def get_chunk(method, chunk, get_peer=False):
            self.assertEqual(method, 'blockchain.block.get_chunk')
            bitcoind_header = "00000020fe52010fa7c798b97621508b772142dfc7b594df7a3a3200000000000000000097db2dc9" \
                              "4e2799bcdd7259f0876467b379fe44b69342df38a8ec2f350722f9a73367a95aa38955173732b883"
            i = {1: 2016, 2: 1024}
            return bitcoind_header * i[chunk]

        self.connectionpool.call.side_effect = get_chunk
        res = self.loop.run_until_complete(self.sut.get_headers_in_range_from_chunks(1, 3))
        Mock.assert_has_calls(
            self.connectionpool.call,
            calls=[call('blockchain.block.get_chunk', 1, get_peer=False),
                   call('blockchain.block.get_chunk', 2, get_peer=False)],
            any_order=True
        )
        i = 0
        for header in res:
            self.assertEqual(header.get('block_height'), i+2016)
            header['block_height'] = 513526
            i += 1
            header_from_chunk = {
                'bits': 391481763,
                'block_hash': '0000000000000000001e9a7d5bdb53bf19487d50d9e68e04b7254662d693a6ea',
                'block_height': 513526,
                'header_bytes': b'\x00\x00\x00 \xfeR\x01\x0f\xa7\xc7\x98\xb9v!P\x8bw!B\xdf'
                b'\xc7\xb5\x94\xdfz:2\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                b"\x97\xdb-\xc9N'\x99\xbc\xddrY\xf0\x87dg\xb3y\xfeD\xb6"
                b'\x93B\xdf8\xa8\xec/5\x07"\xf9\xa73g\xa9Z\xa3\x89U\x17'
                b'72\xb8\x83',
                'merkle_root': 'a7f92207352feca838df4293b644fe79b3676487f05972ddbc99274ec92ddb97',
                'nonce': 2209886775,
                'prev_block_hash': '000000000000000000323a7adf94b5c7df4221778b502176b998c7a70f0152fe',
                'timestamp': 1521051443,
                'version': 536870912
            }
            self.assertEqual(header, header_from_chunk, msg='%s %s' % (header_from_chunk, header))
        self.assertEqual(len(res), 3040)

    def test_get_chunks_no_chunk(self):
        self.connectionpool.call.return_value = async_coro(None)
        self.assertIsNone(self.loop.run_until_complete(self.sut.get_headers_from_chunk(1)))

    def test_get_headers_in_range_ok(self):
        peer = Mock()
        self.connectionpool.call.side_effect = [
            async_coro((peer, self.electrum_header)),
            async_coro((peer, self.electrum_header))
        ]
        res = self.loop.run_until_complete(self.sut.get_headers_in_range(1, 3))
        self.assertEqual(res, [self.parsed_header, self.parsed_header])



import os
import binascii
from spruned.application.context import ctx

TESTING = os.getenv('TESTING')
CHECK_NETWORK_HOST = [
    'a.root-servers.net',
    'b.root-servers.net',
    'c.root-servers.net',
    'd.root-servers.net',
    'e.root-servers.net',
    'f.root-servers.net',
    'g.root-servers.net',
    'h.root-servers.net',
    'i.root-servers.net',
    'j.root-servers.net',
    'k.root-servers.net',
    'l.root-servers.net',
    'm.root-servers.net',
    'www.bitcoin.org'
]
SQLITE_DBNAME = ''
LEVELDB_BLOCKCHAIN_ADDRESS = '/tmp/{}-test.ldb'.format(binascii.hexlify(os.urandom(8)))
LEVELDB_BLOCKCHAIN_SLUG = b'blockchain'
LEVELDB_CACHE_SLUG = b'cache'

VERSION = "0.0.1"
BITCOIND_API_VERSION = "0.16"

if not TESTING:
    STORAGE_ADDRESS = '%s/storage/' % ctx.datadir
    LOGFILE = '%s/spruned.log' % ctx.datadir
    SQLITE_DBNAME = '%sheaders.db' % STORAGE_ADDRESS
    LEVELDB_BLOCKCHAIN_ADDRESS = '%sdatabase.ldb' % STORAGE_ADDRESS



import asyncio
import hashlib
import binascii
from bitcoin import deserialize, serialize, decode, bin_sha256, encode
from spruned.application import exceptions


def normalize_transaction(tx):
    _tx = deserialize(tx)
    _tx['segwit'] = True
    for i, vin in enumerate(_tx['ins']):
        if vin.get('txinwitness', '0'*64) == '0'*64:
            _tx['ins'][i]['txinwitness'] = ''
    return serialize(_tx)


def blockheader_to_blockhash(header: (bytes, str)) -> (bytes, str):
    if isinstance(header, bytes):
        h, fmt = header, 'bin'
    else:
        h, fmt = binascii.unhexlify(header.encode()), 'hex'
    blockhash = hashlib.sha256(hashlib.sha256(h).digest())
    bytes_blockhash = blockhash.digest()[::-1]
    return fmt == 'hex' and binascii.hexlify(bytes_blockhash).decode() or bytes_blockhash


def deserialize_header(header: (str, bytes)):
    if isinstance(header, bytes):
        h, fmt = header, 'bin'
    else:
        h, fmt = binascii.unhexlify(header.encode()), 'hex'
    blockhash = bin_sha256(bin_sha256(h))[::-1]
    data = {
        "version": decode(h[:4][::-1], 256),
        "prev_block_hash": h[4:36][::-1],
        "merkle_root": h[36:68][::-1],
        "timestamp": decode(h[68:72][::-1], 256),
        "bits": decode(h[72:76][::-1], 256),
        "nonce": decode(h[76:80][::-1], 256),
        "hash": blockhash
    }
    if fmt == 'hex':
        data['prev_block_hash'] = binascii.hexlify(data['prev_block_hash']).decode()
        data['merkle_root'] = binascii.hexlify(data['merkle_root']).decode()
        data['hash'] = binascii.hexlify(data['hash']).decode()
    verify_pow(h, blockhash)
    return data


def verify_pow(header, blockhash):
    bits = header[72:76][::-1]
    target = int.from_bytes(bits[1:], 'big') * 2 ** (8 * (bits[0] - 3))
    if target < int.from_bytes(blockhash, 'little'):
        return True
    raise exceptions.InvalidPOWException


def serialize_header(inp):
    o = encode(inp['version'], 256, 4)[::-1] + \
        binascii.unhexlify(inp['prev_block_hash'])[::-1] + \
        binascii.unhexlify(inp['merkle_root'])[::-1] + \
        encode(inp['timestamp'], 256, 4)[::-1] + \
        encode(inp['bits'], 256, 4)[::-1] + \
        encode(inp['nonce'], 256, 4)[::-1]
    h = binascii.hexlify(bin_sha256(bin_sha256(o))[::-1]).decode()
    if inp.get('hash'):
        assert h == inp['hash'], (hashlib.sha256(o), inp['hash'])
    return binascii.hexlify(o).decode()


def get_nearest_parent(number: int, divisor: int):
    return int(number - number % divisor)


async def async_delayed_task(task, seconds: int=0, disable_log=True):
    from spruned.application.logging_factory import Logger
    not disable_log and Logger.root.debug('Scheduling task %s in %s seconds', task, seconds)
    await asyncio.sleep(seconds)
    return await task


def create_directory():
    from spruned import settings
    from spruned.application.context import ctx
    import os
    if not os.path.exists(ctx.datadir):
        os.makedirs(ctx.datadir)
    if not os.path.exists(settings.STORAGE_ADDRESS):
        os.makedirs(settings.STORAGE_ADDRESS)


def check_internet_connection():
    from spruned.application.logging_factory import Logger
    from spruned.settings import CHECK_NETWORK_HOST
    import subprocess
    import os
    Logger.electrum.debug('Checking internet connectivity')
    i = 0
    while i < 10:
        import random
        host = random.choice(CHECK_NETWORK_HOST)
        ret_code = subprocess.call(['ping', '-c', '1', '-W', '5', host],
                                   stdout=open(os.devnull, 'w'),
                                   stderr=open(os.devnull, 'w'))
        if not ret_code:
            return True
        i += 1
    Logger.electrum.debug('No internet connectivity!')
    return False


def script_to_scripthash(script):
    h = hashlib.sha256(bytes.fromhex(script)).digest()[0:32]
    return binascii.hexlify(bytes(reversed(h))).decode('ascii')



import asyncio
import unittest
from unittest.mock import Mock

import time

from spruned.daemon import exceptions
from spruned.daemon.electrod.electrod_connection import ElectrodConnectionPool
from test.utils import async_coro


async def connect(m):
    m._connect.return_value = True
    m.connected = True
    m.protocol = 'protocol'
    return m._connect()


async def disconnect(m):
    m._disconnect.return_value = True
    m.connected = False
    m.protocol = None
    return m._disconnect()


class TestElectrodConnectionPool(unittest.TestCase):
    def setUp(self):
        self.servers = [['hostname{}'.format(x), 's'] for x in range(0, 10)]
        self.electrod_loop = Mock()
        self.delayer = Mock()
        self.electrod_loop = Mock()
        self.network_checker = Mock()
        self.connection_factory = Mock()
        self.sut = ElectrodConnectionPool(
            connections=3,
            loop=self.electrod_loop,
            delayer=self.delayer,
            peers=self.servers,
            network_checker=self.network_checker,
            connection_factory=self.connection_factory
        )
        self.loop = asyncio.get_event_loop()

    async def stop_keepalive(self, n):
        await asyncio.sleep(n)
        self.sut._keepalive = False

    def test_connect_no_internet_connectivity(self):
        self.sut._sleep_on_no_internet_connectivity = 2
        self.network_checker.return_value = False
        self.loop.run_until_complete(
            asyncio.gather(
                self.stop_keepalive(4.1),
                self.sut.connect()
            )
        )
        self.sut._sleep_on_no_internet_connectivity = 30

        self.assertEqual(0, len(self.sut.connections))
        self.assertEqual(4, self.network_checker.call_count)

    def test_connect_success(self):
        self.sut.loop = self.loop
        self.network_checker.return_value = True
        conn1 = Mock(connected=False, score=0)
        conn1.connect = lambda: connect(conn1)
        conn2 = Mock(connected=False, score=0)
        conn2.connect = lambda: connect(conn2)
        conn3 = Mock(connected=False, score=0)
        conn3.connect = lambda: connect(conn3)
        self.connection_factory.side_effect = [conn1, conn2, conn3]
        on_connected_observer = Mock()
        on_connected_observer.return_value = async_coro(True)
        self.sut.add_on_connected_observer(on_connected_observer)
        self.loop.run_until_complete(
            asyncio.gather(
                self.stop_keepalive(7),
                self.sut.connect()
            )
        )
        self.sut.loop = self.electrod_loop

        for conn in [conn1, conn2, conn3]:
            self.assertEqual(conn.add_on_connect_callback.call_count, 1)
            self.assertEqual(conn.add_on_header_callbacks.call_count, 1)
            self.assertEqual(conn.add_on_peers_callback.call_count, 1)
            self.assertEqual(conn.add_on_error_callback.call_count, 1)
            self.assertTrue(conn.connected)
        self.assertEqual(on_connected_observer.call_count, 1)
        self.assertEqual(3, len(self.sut.connections))
        self.assertEqual(3, len(self.sut.established_connections))

    def test_connect_too_much_peers(self):
        """
        I didn't tracked with an issue when this happened,
        but it did, maybe asyncio lag\race conditions
        """
        self.sut.loop = self.loop
        self.network_checker.return_value = True
        self.sut._required_connections = 2
        conn1 = Mock(score=10, connected=False, start_score=10)
        conn1.connect = lambda: connect(conn1)
        conn1.disconnect = lambda: disconnect(conn1)
        conn2 = Mock(score=10, connected=False, start_score=10)
        conn1.score = 10
        conn2.connect = lambda: connect(conn2)
        conn2.disconnect = lambda: disconnect(conn2)
        self.connection_factory.side_effect = [conn1, conn2]
        on_connected_observer = Mock()
        on_connected_observer.return_value = async_coro(True)
        self.sut.add_on_connected_observer(on_connected_observer)

        async def _change_behaviour(n):
            await asyncio.sleep(n)
            self.assertEqual(len(self.sut.connections), 2)
            self.sut._required_connections = 1

        self.loop.run_until_complete(
            asyncio.gather(
                self.stop_keepalive(15),
                _change_behaviour(6),
                self.sut.connect()
            )
        )

        self.assertEqual(1, len(self.sut.connections))
        self.assertEqual(1, len(self.sut.established_connections))
        d = 0
        c = 0
        for conn in [conn1, conn2]:
            c += conn._connect.call_count
            d += conn._disconnect.call_count
        self.assertEqual(d, 1)
        self.assertEqual(c, 2)
        self.assertEqual(1, len([c for c in [conn1, conn2] if c.connected]))

    def test_call_corners(self):
        with self.assertRaises(ValueError):
            self.loop.run_until_complete(self.sut.call('cafe', {'par': 'ams'}, get_peer=True, agreement=2))
        with self.assertRaises(ValueError):
            self.sut._required_connections = 2
            self.loop.run_until_complete(
                self.sut.call('cafe', {'par': 'ams'}, agreement=self.sut._required_connections+1)
            )
        with self.assertRaises(exceptions.NoPeersException):
            self.loop.run_until_complete(self.sut.call('cafe', 'babe'))

    def test_call_success(self):
        response = 'some response'
        conn = Mock(connected=True, protocol=True, score=10)
        self.sut._connections = [conn]
        conn.rpc_call.return_value = async_coro(response)
        res = self.loop.run_until_complete(self.sut.call('cafe', 'babe'))
        self.assertEqual(res, response)

        conn.rpc_call.return_value = async_coro(response)
        peer, res = self.loop.run_until_complete(self.sut.call('cafe', 'babe', get_peer=True))
        self.assertEqual(peer, conn)
        self.assertEqual(res, response)

    def test_call_success_multiple_agreement(self):
        response = 'some response'
        conn = Mock(connected=True, protocol=True, score=10)
        conn.rpc_call.return_value = async_coro(response)
        conn2 = Mock(connected=True, protocol=True, score=10)
        conn2.rpc_call.return_value = async_coro(response)
        self.sut._connections = [conn, conn2]
        res = self.loop.run_until_complete(self.sut.call('cafe', 'babe', agreement=2))
        self.assertEqual(res, response)

    def test_call_failure_not_enough_responses(self):
        response = 'some response'
        conn = Mock(connected=True, protocol=True, score=10)
        conn.rpc_call.return_value = async_coro(response)
        conn2 = Mock(connected=True, protocol=True, score=10)
        conn2.rpc_call.return_value = async_coro(None)
        self.sut._connections = [conn, conn2]
        with self.assertRaises(exceptions.ElectrodMissingResponseException):
            self.loop.run_until_complete(self.sut.call('cafe', 'babe', agreement=2))

        conn.rpc_call.return_value = async_coro(response)
        conn2.rpc_call.return_value = async_coro(None)
        self.assertIsNone(
            self.loop.run_until_complete(
                self.sut.call('cafe', 'babe', agreement=2, fail_silent=True)
            )
        )

    def test_call_failure_disagreement_on_responses(self):
        response = 'some response'
        response2 = 'another response'
        conn = Mock(connected=True, protocol=True, score=10)
        conn.rpc_call.return_value = async_coro(response)
        conn2 = Mock(connected=True, protocol=True, score=10)
        conn2.rpc_call.return_value = async_coro(response2)
        self.sut._connections = [conn, conn2]
        with self.assertRaises(exceptions.NoQuorumOnResponsesException):
            self.loop.run_until_complete(self.sut.call('cafe', 'babe', agreement=2))

        conn.rpc_call.return_value = async_coro(response)
        conn2.rpc_call.return_value = async_coro(response2)
        self.assertIsNone(
            self.loop.run_until_complete(
                self.sut.call('cafe', 'babe', agreement=2, fail_silent=True)
            )
        )

    def test_call_missing_response(self):
        conn = Mock(connected=True, protocol=True, score=10)
        conn2 = Mock(connected=True, protocol=True, score=10)
        conn.rpc_call.return_value = async_coro(None)
        conn2.rpc_call.return_value = async_coro(None)
        self.sut._connections = [conn, conn2]
        with self.assertRaises(exceptions.ElectrodMissingResponseException):
            self.loop.run_until_complete(
                self.sut.call('cafe', 'babe', agreement=2)
            )

    def test_corners(self):
        s = [s for s in self.sut._peers]
        self.sut._peers = []
        with self.assertRaises(exceptions.NoServersException):
            self.sut._pick_peer()
        with self.assertRaises(exceptions.NoServersException):
            self.sut._pick_multiple_peers(2)
        self.sut._peers = s
        with self.assertRaises(exceptions.NoPeersException):
            self.sut._pick_connection()
        with self.assertRaises(exceptions.NoPeersException):
            self.sut._pick_multiple_connections(2)
        self.assertIsNone(
            self.sut._pick_connection(fail_silent=True)
        )
        self.sut._peers = ['cafebabe']
        self.sut._connections.append(Mock(hostname='cafebabe', connected=True, score=1))
        with self.assertRaises(exceptions.NoServersException):
            self.sut._pick_peer()
        with self.assertRaises(exceptions.NoServersException):
            self.sut._pick_multiple_peers(1)

    def test__handle_peer_error_disconnected(self):
        conn = Mock(connected=False)
        res = self.loop.run_until_complete(self.sut._handle_peer_error(conn))
        self.assertIsNone(res)

    def test_handle_peer_error_noscore(self):
        conn = Mock(connected=True, score=0)
        conn.disconnect.return_value = 'disconnect'
        self.delayer.return_value = 'delayer'
        res = self.loop.run_until_complete(self.sut._handle_peer_error(conn))
        self.assertIsNone(res)
        Mock.assert_called_with(self.electrod_loop.create_task, 'delayer')
        Mock.assert_called_with(self.delayer, 'disconnect')

    def test_handle_peer_error_ping_timeout(self):
        conn = Mock(connected=True, score=10)
        conn.ping.return_value = async_coro(None)
        conn.disconnect.return_value = 'disconnect'
        self.delayer.return_value = 'delayer'

        res = self.loop.run_until_complete(self.sut._handle_peer_error(conn))
        self.assertIsNone(res)
        Mock.assert_called_with(self.electrod_loop.create_task, 'delayer')
        Mock.assert_called_with(self.delayer, 'disconnect')

    def test_on_connected_callback(self):
        peer = Mock()
        cob = Mock()
        self.network_checker.return_value = True
        self.sut.add_on_connected_observer(cob)
        peer.subscribe.return_value = 'subscriberrr'
        self.delayer.return_value = 'delayerrr'
        self.loop.run_until_complete(self.sut.on_peer_connected(peer))
        Mock.assert_called_with(
            peer.subscribe,
            'blockchain.headers.subscribe',
            self.sut.on_peer_received_header,
            self.sut.on_peer_received_header,
        )
        Mock.assert_called_once_with(self.electrod_loop.create_task, 'delayerrr')
        Mock.assert_called_once_with(self.delayer, 'subscriberrr')

    def test_on_headers_callback(self):
        peer = Mock(last_header='header')
        hob = Mock()
        hob.return_value = 'observing'
        self.delayer.return_value = 'delayerr'
        self.network_checker.return_value = True
        self.sut.add_header_observer(hob)
        self.loop.run_until_complete(self.sut.on_peer_received_header(peer))
        Mock.assert_called_once_with(self.electrod_loop.create_task, 'delayerr')
        Mock.assert_called_once_with(hob, peer, 'header')
        Mock.assert_called_once_with(self.delayer, 'observing')

    def test_on_peer_error(self):
        peer = Mock(is_online=True, connected=False, _errors=[])
        peer.add_error = lambda *x: peer._errors.append(x[0]) if x else peer._errors.append(int(time.time()))
        self.loop.run_until_complete(self.sut.on_peer_error(peer))
        self.assertEqual(len(peer._errors), 1)

    def test_on_peer_error_during_connection(self):
        peer = Mock(is_online=True, connected=False, _errors=[])
        peer.add_error = lambda *x: peer._errors.append(x[0]) if x else peer._errors.append(int(time.time()))
        self.network_checker.return_value = False
        self.loop.run_until_complete(self.sut.on_peer_error(peer, error_type='connect'))
        self.assertEqual(len(peer._errors), 0)

        self.network_checker.return_value = True
        self.loop.run_until_complete(self.sut.on_peer_error(peer, error_type='connect'))
        self.assertEqual(len(peer._errors), 1)